#include <windows.h>
#include <math.h>
#ifndef M_PI
#define	M_PI  3.14159265358979323846
#endif
#include <stdio.h>
#include "resource.h"

#define NAXE 3
HWND hDlgMvt = NULL;				// hwnd de la bote de dialogue
static WNDPROC StdEditMvtProc;		// Memo procedure standard du controle d'dition
static HWND hEdit[NAXE];

#define VALIDE	500				// IDC bouton virtuel valide
#define ANNULE	501				// IDC bouton virtuel annule


/* Procedure interceptant les entres au control-edit IDC_EDIT1 pour simuler les
   action valide/annule avec RETURN/ESCAPE */
LRESULT CALLBACK MvtEditProc (HWND hWnd, UINT iMsg, WPARAM wParam, LPARAM lParam)
{
	/* Si reoit un return ou un escape,  on simule l'appui sur 
       des boutons virtuels valide ou annule */
    if (iMsg == WM_CHAR)
	{
		if(wParam == VK_RETURN)
		{
			SendMessage (hDlgMvt, WM_COMMAND, MAKELONG(VALIDE,0), (LPARAM) hWnd);
			return 0;
		}
		if (wParam == VK_ESCAPE)
		{
			SendMessage (hDlgMvt, WM_COMMAND, MAKELONG(ANNULE,0), (LPARAM) hWnd); 
			return 0;
		}
	}

    return CallWindowProc (StdEditMvtProc, hWnd, iMsg, wParam, lParam) ;
}

static int idc_edit[NAXE] = {ID2_EDIT1, ID2_EDIT2, ID2_EDIT3};

static int idc_static[NAXE] = {IDC_STATIC1, IDC_STATIC2, IDC_STATIC3};

static int idc_scroll[NAXE] = {IDC_SCROLLBAR1, IDC_SCROLLBAR2, IDC_SCROLLBAR3};

static double Valmin[NAXE] = {-10., -10., -180.};
static double Valmax[NAXE] = { 10.,  10.,  180.};
static double val_mem[NAXE];
extern double xyzltr[];
void UpdateRov(int fmodconf, int fmodvue, int fmodsit);

/* Call back traitant les messages de la bote de dialogue */
BOOL CALLBACK DialogMvt(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam) 
{
#define MAXSCROLL 1000
static HWND hScroll[NAXE];
static int ix_mem[NAXE];
int ix, numax;
#define MAXCHAR1 20
char text[MAXCHAR1+1]="";
int n, lwp;

    switch (message) 
	{ 
	case WM_INITDIALOG :
//#define MODALE_cad_BLOQUE
#ifdef MODALE_cad_BLOQUE
		hDlgMvt = hDlg;
#endif
		/* On sous-classe le control-edit*/
		for (n=0; n<NAXE; n++)
		{
			hEdit[n] = GetDlgItem (hDlg, idc_edit[n]);
			StdEditMvtProc = (WNDPROC) SetWindowLong(hEdit[n], GWL_WNDPROC, (LONG) MvtEditProc);
			/* On affiche les valeurs initiales  l'afficheur et  l'editeur */
			if (n == 2) val_mem[n] = xyzltr[3]*180./M_PI;
			else val_mem[n] = xyzltr[n];
			sprintf(text,"%5.2lf",val_mem[n]);
			SetDlgItemText(hDlg, idc_edit[n], text);
			SetDlgItemText(hDlg, idc_static[n], text);
			/* On initialise le curseur */
			hScroll[n] = GetDlgItem (hDlg, idc_scroll[n]);
			SetScrollRange (hScroll[n] , SB_CTL, 0, MAXSCROLL, FALSE) ;
			/* On positionne le curseur  la valeur initiale */
			ix_mem[n] = (int) (MAXSCROLL*(val_mem[n] - Valmin[n])/(Valmax[n]-Valmin[n]));
			SetScrollPos (hScroll[n], SB_CTL, ix_mem[n], TRUE) ;
		}
		return TRUE ;

    case WM_COMMAND :
		numax = -1;
		for (n=0; n<NAXE; n++) if(hEdit[n] == (HWND) lParam) { numax = n; break;}
		lwp = LOWORD(wParam);
        switch (lwp) 
		{ 
		case VALIDE : /* Valide dans fenetre edit */
			if (numax < 0) break;
			/* lecture texte du controle edition */
			n = GetDlgItemText(hDlg, idc_edit[numax], text, MAXCHAR1);
			/* test des valeurs / aux limites */
			if (n)
			{
				n = sscanf(text,"%lf",&val_mem[numax]);
				if (n!=1) val_mem[numax] = 0.;
				if (val_mem[numax] < Valmin[numax])
				{
					val_mem[numax] = Valmin[numax];
					MessageBeep(0xFFFFFFFF);
				}
				if (val_mem[numax] > Valmax[numax])
				{
					val_mem[numax] = Valmax[numax];
					MessageBeep(0xFFFFFFFF);
				}
			}
			/* On va afficher la valeur et mettre le curseur en place */
			ix_mem[numax] = (int) (MAXSCROLL*(val_mem[numax] - Valmin[numax])/(Valmax[numax]-Valmin[numax]));
		case ANNULE : /* annule dans fenetre edit, ou fin du valide */
			/* On affiche et on positionne le curseur avec les valeurs _mem */
			if (numax < 0) break;
			sprintf(text,"%5.2lf",val_mem[numax]);
			SetDlgItemText(hDlg, idc_edit[numax], text);
			SetDlgItemText(hDlg, idc_static[numax], text);
			SetScrollPos (hScroll[numax], SB_CTL, ix_mem[numax], TRUE) ;
			if (numax == 2) xyzltr[3] = val_mem[numax]*M_PI/180.;
			else xyzltr[numax] = val_mem[numax];
			if (lwp == VALIDE) UpdateRov(0,0,1);
			return TRUE;

		case IDCANCEL :
#ifdef MODALE_cad_BLOQUE
			EndDialog (hDlg, 0) ;
#else
			DestroyWindow(hDlg);
#endif
			hDlgMvt = NULL;
//		    PostQuitMessage (0) ; // On envoie le WM_QUIT- On TUE l'application
			return TRUE ;
        } /* Fin du switch (LOWORD(wParam)) dans WM_COMMAND */
		break ; /* Fin du WM_COMMAND */

	case WM_VSCROLL :
	case WM_HSCROLL :
		numax = -1;
		for (n=0; n<NAXE; n++) if(hScroll[n] == (HWND) lParam) { numax = n; break;}
		if (numax < 0) return FALSE ;
		ix = ix_mem[numax];
		switch (LOWORD(wParam))
		{
		case SB_PAGEDOWN : 
			ix += MAXSCROLL/20 ;
			break ;
		case SB_LINEDOWN :
			++ix ;
			break ;
		case SB_PAGEUP : 
			ix -= MAXSCROLL/20 ;
			break ;
		case SB_LINEUP :
			--ix ;
			break ;
		case SB_TOP :
			ix = 0 ;
			break ;
		case SB_BOTTOM :
			ix = MAXSCROLL ;
			break ;
		case SB_THUMBPOSITION :
		case SB_THUMBTRACK :
			ix = HIWORD (wParam) ;
			break ;
		default :
			return FALSE ;
		} /* Fin du switch (LOWORD(wParam)) dans WM_VSCROLL */

		ix = ix < MAXSCROLL ? ix : MAXSCROLL ;
		ix = ix > 0 ? ix : 0 ;

		if (ix != ix_mem[numax])
		{
				ix_mem[numax] = ix ;
				SetScrollPos (hScroll[numax], SB_CTL, ix, TRUE) ;
				val_mem[numax] = Valmin[numax] + (Valmax[numax]-Valmin[numax])*ix_mem[numax]/MAXSCROLL;
				sprintf(text,"%5.2lf",val_mem[numax]);
				SetDlgItemText(hDlg, idc_edit[numax], text);
				SetDlgItemText(hDlg, idc_static[numax], text);
				if (numax == 2) xyzltr[3] = val_mem[numax]*M_PI/180.;
				else xyzltr[numax] = val_mem[numax];
				UpdateRov(0,0,1);
		}
		return TRUE ; /* Fin du WM_VSCROLL */
	} /* Fin du switch (message) */

    return FALSE; 
} 

#if 0
int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
                    PSTR szCmdLine, int iCmdShow)
{ 
MSG  msg ;


	hDlgMvt = CreateDialog(hInstance, MAKEINTRESOURCE(IDD_DIALOG1), (HWND) NULL, 
			(DLGPROC)DialogConfRov) ;
	ShowWindow (hDlgMvt, SW_SHOW) ;


	while (GetMessage (&msg, NULL, 0, 0))
	{ 
		TranslateMessage (&msg) ;	
		DispatchMessage (&msg) ;
	}
	return msg.wParam ;
}
#endif