#include <math.h>
#ifndef M_PI
#define	M_PI  3.14159265358979323846
#endif
#include "geouti.h"

#define GLOBAL
GLOBAL double conf_rov[21] ;
GLOBAL double xyzltr[6];


#define SATUR(x,a) if(x > (a)) x=(a) ; else if(x < -(a)) x=-(a) ;
#define R2DEG *180./M_PI
#define D2RAD *M_PI/180.
#define cte_pneu   0.35
#define cte_a      0.2
#define cte_d      0.2
#define cte_b      0.9
#define cte_c      0.7
#define cte_f      0.6
#define cte_e      0.2
#define cte_g      0.10
#define cte_h      0.
#define cte_xh     0.2
#define cte_zh     0.3
#define cte_l      0.4
#define cte_m      0.4
#define cte_rho    0.5
static double cte_cpneu, cte_spneu, cte_carros, cte_inclin, cte_ipc,
    sin_ipc, cos_ipc, sin_car, cos_car, cte_b2, cte_c2, cte_2bc, cte_l2,
    cte_2m, cte_m2, cte_GR, cte_s, cte_t, cte_v, cte_rsp, cte_rcp ;
static double eta_d0 = 0., eta_g0 = 0.;
static double theta_max ;

static double
    UGd[3], UGg[3], URd[3], URg[3],
    D_TS[12], D_TAa[12], D_TAr[12],
    D_ABa[12], D_ABr[12], D_BCa[12], D_BCr[12], D_CSa[12], D_CSr[12],
    D_SPad[12], D_SPag[12], D_SPcd[12], D_SPcg[12], D_SPrd[12], D_SPrg[12],
    D_PGad[12], D_PGag[12], D_PGcd[12], D_PGcg[12], D_PGrd[12], D_PGrg[12],
    GRd[3], GRg[3] ;

static double ORad[3], ORag[3], ORcd[3], ORcg[3], ORrd[3], ORrg[3] ;
static double OQad[3], OQag[3], OQcd[3], OQcg[3], OQrd[3], OQrg[3] ;
static double D_OT[12] ;
static double
    D_OSc[12],
    D_OAa[12], D_OBa[12], D_OCa[12], D_OSa[12],
    D_OAr[12], D_OBr[12], D_OCr[12], D_OSr[12],
    D_OPad[12], D_OGad[12], D_OPag[12], D_OGag[12],
    D_OPcd[12], D_OGcd[12], D_OPcg[12], D_OGcg[12],
    D_OPrd[12], D_OGrd[12], D_OPrg[12], D_OGrg[12] ;

/*--------------------------------------------------------------------------*/
int theta2eta (double *eta_d, double *eta_g, double theta)
/*--------------------------------------------------------------------------*/
{
int satur  = 0 ;
double
    ct, st, X, Y, n, xi, ckimxi, kimxi, ki_d, ki_g ;

    ct = cos (theta) ; st = sin (theta) ;
    X =  cte_f * ct - cte_e * st - cte_xh ;
    Y = -cte_f * st - cte_e * ct - cte_zh ;
    n = sqrt (X*X + Y*Y) ; xi = atan2 (-Y,X) ;
    ckimxi = (cte_l2 - (n*n + cte_m2))/(cte_2m*n) ;
    if (ckimxi > 1.) {ckimxi = 1. ; satur = 1 ;}
    else if (ckimxi < -1.) {ckimxi = -1. ; satur = 1 ;}
    kimxi = - acos (ckimxi) ; /* Ngatif a droite */
    ki_d = kimxi + xi ;
    *eta_d = ki_d - eta_d0 - theta ;

    X = -cte_f * ct - cte_e * st + cte_xh ;
    Y =  cte_f * st - cte_e * ct - cte_zh ;
    n = sqrt (X*X + Y*Y) ; xi = atan2 (-Y,X) ;
    ckimxi = (cte_l2 - (n*n + cte_m2))/(cte_2m*n) ;
    if (ckimxi > 1.) {ckimxi = 1. ; satur = 1 ;}
    else if (ckimxi < -1.) {ckimxi = -1. ; satur = 1 ;}
    kimxi = acos (ckimxi) ;  /* Positif a gauche */
    ki_g = kimxi + xi - 2. * M_PI ;
    *eta_g = ki_g - eta_g0 - theta ;
    return satur ;
}


/*--------------------------------------------------------------------------*/
void geodir_rover (double *xyzltr, double *conf)
/*--------------------------------------------------------------------------*/
{
double a, eta_ad, eta_ag, eta_cd, eta_cg, eta_rd, eta_rg ;

    cpyvec (D_OT, xyzltr) ;
    ltrmat (D_OT +3, xyzltr+3) ;

  /* Essieu central */
    mrotna (D_TS +3, 2, conf[7]) ; /* theta_c */
    pordpord (D_OSc, D_OT, D_TS) ;

  /* Essieu avant */
    mrotna (D_TAa+3, 1, conf[0]) ; /* alfa_a */
    pordpord (D_OAa, D_OT, D_TAa) ;
    mrotna (D_ABa+3, 1, conf[2]) ; /* beta_a */
    pordpord (D_OBa, D_OAa, D_ABa) ;
    mrotna (D_BCa+3, 1, conf[4]) ; /* beta_a */
    pordpord (D_OCa, D_OBa, D_BCa) ;
    mrotna (D_CSa+3, 2, conf[6]) ; /* theta_a */
    pordpord (D_OSa, D_OCa, D_CSa) ;

  /* Essieu arriere */
    mrotna (D_TAr+3, 1, conf[1]) ; /* alfa_r */
    pordpord (D_OAr, D_OT, D_TAr) ;
    mrotna (D_ABr+3, 1, conf[3]) ; /* beta_r */
    pordpord (D_OBr, D_OAr, D_ABr) ;
    mrotna (D_BCr+3, 1, conf[5]) ; /* gamma_r */
    pordpord (D_OCr, D_OBr, D_BCr) ;
    mrotna (D_CSr+3, 2, conf[8]) ; /* theta_r */
    pordpord (D_OSr, D_OCr, D_CSr) ;

  /* Roues avant */
    theta2eta (&eta_ad, &eta_ag, conf[6]) ; /* theta_a */
    mrotna (D_SPad+3, 2, eta_ad) ;
    mrotna (D_SPag+3, 2, eta_ag) ;

    pordpord (D_OPad, D_OSa, D_SPad) ;
    mrotva (D_PGad+3, UGd, conf[9]) ;  /* delta_ad */
    pordpord (D_OGad, D_OPad, D_PGad) ;
    pordvec (ORad, D_OGad, GRd) ;

    pordpord (D_OPag, D_OSa, D_SPag) ;
    mrotva (D_PGag+3, UGg, conf[10]) ; /* delta_ag */
    pordpord (D_OGag, D_OPag, D_PGag) ;
    pordvec (ORag, D_OGag, GRg) ;

  /* Roues centre */
    theta2eta (&eta_cd, &eta_cg, conf[7]) ; /* theta_c */
    mrotna (D_SPcd+3, 2, eta_cd) ;
    mrotna (D_SPcg+3, 2, eta_cg) ;

    pordpord (D_OPcd, D_OSc, D_SPcd) ;
    mrotva (D_PGcd+3, UGd, conf[11]) ; /* delta_cd */
    pordpord (D_OGcd, D_OPcd, D_PGcd) ;
    pordvec (ORcd, D_OGcd, GRd) ;

    pordpord (D_OPcg, D_OSc, D_SPcg) ;
    mrotva (D_PGcg+3, UGg, conf[12]) ; /* delta_cg */
    pordpord (D_OGcg, D_OPcg, D_PGcg) ;
    pordvec (ORcg, D_OGcg, GRg) ;

  /* Roues arriere */
    theta2eta (&eta_rd, &eta_rg, conf[8]) ; /* theta_r */
    mrotna (D_SPrd+3, 2, eta_rd) ;
    mrotna (D_SPrg+3, 2, eta_rg) ;

    pordpord (D_OPrd, D_OSr, D_SPrd) ;
    mrotva (D_PGrd+3, UGd, conf[13]) ; /* delta_rd */
    pordpord (D_OGrd, D_OPrd, D_PGrd) ;
    pordvec (ORrd, D_OGrd, GRd) ;

    pordpord (D_OPrg, D_OSr, D_SPrg) ;
    mrotva (D_PGrg+3, UGg, conf[14]) ; /* delta_rg */
    pordpord (D_OGrg, D_OPrg, D_PGrg) ;
    pordvec (ORrg, D_OGrg, GRg) ;

/* Pour le graphique */
    a = eta_ad + eta_d0 + conf[6] ;
    OQad[0] = D_OPad[0] + cos(a)*cte_m ;
    OQad[1] = D_OPad[1] ;
    OQad[2] = D_OPad[2] - sin(a)*cte_m ;
    a = eta_ag + eta_g0 + conf[6] ;
    OQag[0] = D_OPag[0] + cos(a)*cte_m ;
    OQag[1] = D_OPag[1] ;
    OQag[2] = D_OPag[2] - sin(a)*cte_m ;
    a = eta_cd + eta_d0 + conf[7] ;
    OQcd[0] = D_OPcd[0] + cos(a)*cte_m ;
    OQcd[1] = D_OPcd[1] ;
    OQcd[2] = D_OPcd[2] - sin(a)*cte_m ;
    a = eta_cg + eta_g0 + conf[7] ;
    OQcg[0] = D_OPcg[0] + cos(a)*cte_m ;
    OQcg[1] = D_OPcg[1] ;
    OQcg[2] = D_OPcg[2] - sin(a)*cte_m ;
    a = eta_rd + eta_d0 + conf[8] ;
    OQrd[0] = D_OPrd[0] + cos(a)*cte_m ;
    OQrd[1] = D_OPrd[1] ;
    OQrd[2] = D_OPrd[2] - sin(a)*cte_m ;
    a = eta_rg + eta_g0 + conf[8] ;
    OQrg[0] = D_OPrg[0] + cos(a)*cte_m ;
    OQrg[1] = D_OPrg[1] ;
    OQrg[2] = D_OPrg[2] - sin(a)*cte_m ;
}

#define TERRAIN_C
#ifndef TERRAIN_C
double h2sol (double *xyz, double R) { return xyz[2] - R ;}
#else
double h2sol (double *xyz, double R);
#endif
/*--------------------------------------------------------------------------*/
int geoinv_rover (double *xyzltr, double *conf, double tad, double trd,
								double E, int imax)
/*--------------------------------------------------------------------------*/
/* Attention suppose qu'un geodir prealable a ete effectue !!!!!!!!!!!!!!!!!*/
/* Un Geodir final est effectue pour mettre en concordance les conf et xyz  */
/* Le prog calcule xyzltr[2] = z,                                           */
/*                 conf[0]=alfa_a , conf[1]=alfa_r,                         */
/*		   conf[4]=gamma_a, conf[5]=gamma_r,								*/
/*		   conf[6]=theta_a, conf[7]=theta_c, conf[8]=theta_r				*/
/* a partir de la donnee des autres xyzltr, des autres conf, de :           */
/* tad = angle de tangage avant                                             */
/* trd = angle de tangage arriere                                           */
/* Il renvoi le nombre d'iteration effectuees pour atteidre la precision E  */
/* qui est limite a imax                                                    */
/*--------------------------------------------------------------------------*/
{
int iter = 0 ;
double daa, dar, dzad, dzag, dzcd, dzcg, dzrd, dzrg, dtaa, dtar ;
double eps, La, Lr ;
double R = cte_rho  + E;

  do
  {
    La = sqrt ( cte_b2 + cte_c2 + cte_2bc * cos (conf[2])) + 2.*cte_a ;
    Lr = sqrt ( cte_b2 + cte_c2 + cte_2bc * cos (conf[3])) + 2.*cte_a ;

    eps = 0. ;
    dzad = h2sol(ORad, R) ; if (fabs (dzad) > eps) eps = fabs(dzad) ;
    dzag = h2sol(ORag, R) ; if (fabs (dzag) > eps) eps = fabs(dzag) ;
    dzcd = h2sol(ORcd, R) ; if (fabs (dzcd) > eps) eps = fabs(dzcd) ;
    dzcg = h2sol(ORcg, R) ; if (fabs (dzcg) > eps) eps = fabs(dzcg) ;
    dzrd = h2sol(ORrd, R) ; if (fabs (dzrd) > eps) eps = fabs(dzrd) ;
    dzrg = h2sol(ORrg, R) ; if (fabs (dzrg) > eps) eps = fabs(dzrg) ;

    dtaa = tad-asin (*(D_OCa+8)) ; if (fabs (dtaa) > eps) eps = fabs(dtaa) ;
    dtar = trd-asin (*(D_OCr+8)) ; if (fabs (dtar) > eps) eps = fabs(dtar) ;
    xyzltr[2] -= (dzcd + dzcg) / 2. ;
    daa = -(dzad + dzag) / (2.*La) ;
    dar =  (dzrd + dzrg) / (2.*Lr) ;
    conf[0] += daa ;                   /* alfa_a */
    conf[1] += dar ;                   /* alfa_r */
    conf[4] += dtaa - daa ;   /* gamma_a */
    conf[5] += dtar - dar ;   /* gamma_r */
    conf[6] += (dzad - dzag) /cte_v  ; /* theta_a */
    conf[7] += (dzcd - dzcg) /cte_v  ; /* theta_c */
    conf[8] += (dzrd - dzrg) /cte_v  ; /* theta_r */
//    SATUR (conf[6],theta_max) ;
//    SATUR (conf[7],theta_max) ;
//    SATUR (conf[8],theta_max) ;
    geodir_rover (xyzltr, conf) ;
    iter++ ;
  } while (eps > E && iter < imax) ;
  return iter ;
}

void avance(double dist, double courbure)
{
double cLa, cLr, cvs2, upcvs2, umcvs2, dPhi, dLac;
double dLongi, dTransv, cpsi, spsi;

    cLa = courbure * (sqrt ( cte_b2 + cte_c2 + cte_2bc * cos (conf_rov[2])) + 2.*cte_a) ;
    cLr = courbure * (sqrt ( cte_b2 + cte_c2 + cte_2bc * cos (conf_rov[3])) + 2.*cte_a) ;
	cvs2 = courbure * cte_v / 2.;
	upcvs2 = 1. + cvs2;
	umcvs2 = 1. - cvs2;

	conf_rov[9] = atan(cLa/upcvs2);
	conf_rov[10] = atan(cLa/umcvs2);
	conf_rov[11] = 0.;
	conf_rov[12] = 0.;
	conf_rov[13] = -atan(cLr/upcvs2);;
	conf_rov[14] = -atan(cLr/umcvs2);

	dPhi = dist/cte_rho;
	conf_rov[15] += dPhi*upcvs2/cos(conf_rov[ 9]);
	conf_rov[16] += dPhi*umcvs2/cos(conf_rov[10]);
	conf_rov[17] += dPhi*upcvs2/cos(conf_rov[11]);
	conf_rov[18] += dPhi*umcvs2/cos(conf_rov[12]);
	conf_rov[19] += dPhi*upcvs2/cos(conf_rov[13]);
	conf_rov[20] += dPhi*umcvs2/cos(conf_rov[14]);

	dLac = courbure*dist;
	if (courbure == 0.) 
	{
		dLongi = dist;
		dTransv = 0.;
	}
	else
	{
		dLongi = sin(dLac)/courbure;
		dTransv = -(1. - cos(dLac))/courbure;
	}
	cpsi = cos(xyzltr[3]);
	spsi = sin(xyzltr[3]);
	xyzltr[0] += dTransv*cpsi - dLongi*spsi;
	xyzltr[1] += dTransv*spsi + dLongi*cpsi;
	xyzltr[3] += dLac;
    geodir_rover (xyzltr, conf_rov) ;
	geoinv_rover (xyzltr, conf_rov, 0., 0.,
								0.005, 10);
}

/*--------------------------------------------------------------------------*/
void init_rover ()
/*--------------------------------------------------------------------------*/
{
int i ;
double a = 0., da = M_PI/2., eps = 0.00001, e1, e2 ;
/* Carrossage et inclinaisons sont dfinis pour la roue gauche  de telle */
/* maniere que inclin > 0. Les signes sont a inverser pour la roue droite */
    cte_cpneu = cos(cte_pneu) ;
    cte_spneu = sin(cte_pneu) ;
    cte_rsp = cte_rho * cte_spneu ;
    cte_rcp = cte_rho * cte_cpneu ;
    cte_carros = -2.5* M_PI/180. ;
    cte_inclin = 5.* M_PI/180. ;
    cte_ipc = cte_inclin + cte_carros ;
    sin_ipc = sin(cte_ipc) ;
    cos_ipc = cos(cte_ipc) ;
    sin_car = sin(cte_carros) ;
    cos_car = cos(cte_carros) ;
    cte_b2 = cte_b*cte_b ;
    cte_c2 = cte_c*cte_c ;
    cte_2bc = 2.*cte_b*cte_c ;
    cte_l2 = cte_l*cte_l ;
    cte_2m = cte_m+cte_m ;
    cte_m2 = cte_m*cte_m ;
    cte_GR = cte_rho * tan (cte_inclin) ;
    cte_s = cte_GR*cos_car ;
    cte_t = cte_GR*sin_car ;
    cte_v = 2.*(cte_f+cte_g+cte_s) ;

  /* Calcul eta_d0 et eta_g0 */
    i = theta2eta (&eta_d0, &eta_g0, 0.) ;
    while (da > eps)
    {
	da /= 2. ;
	if (i) a -= da ; else a += da ;
	i = theta2eta (&e1, &e2, a) ;
    }
    theta_max = a ;

  /* Axe pivot */
    UGd[0] = -sin_ipc ; UGd[1] = 0.     ; UGd[2] =  cos_ipc ;
    UGg[0] =  sin_ipc ; UGg[1] = 0.     ; UGg[2] =  cos_ipc ;
  /* Axe essieu */
    URd[0] =  cos_car ; URd[1] = 0.     ; URd[2] = sin_car ;
    URg[0] =  cos_car ; URg[1] = 0.     ; URg[2] = -sin_car ;
  /* Tranlations constantes */
    D_TAa[0] = 0.       ; D_TAa[1] =  cte_a ; D_TAa[2] = cte_d   ;
    D_TAr[0] = 0.       ; D_TAr[1] = -cte_a ; D_TAr[2] = cte_d   ;
    D_ABa[0] = 0.       ; D_ABa[1] =  cte_b ; D_ABa[2] = 0.       ;
    D_ABr[0] = 0.       ; D_ABr[1] = -cte_b ; D_ABr[2] = 0.       ;
    D_BCa[0] = 0.       ; D_BCa[1] =  cte_c ; D_BCa[2] = 0.       ;
    D_BCr[0] = 0.       ; D_BCr[1] = -cte_c ; D_BCr[2] = 0.       ;
    D_CSa[0] = 0.       ; D_CSa[1] =  cte_a ; D_CSa[2] = -cte_d    ;
    D_CSr[0] = 0.       ; D_CSr[1] = -cte_a ; D_CSr[2] = -cte_d    ;
    D_SPad[0] =  cte_f  ; D_SPad[1] = 0.    ; D_SPad[2] = -cte_e  ;
    D_SPag[0] = -cte_f  ; D_SPag[1] = 0.    ; D_SPag[2] = -cte_e  ;
    D_SPcd[0] =  cte_f  ; D_SPcd[1] = 0.    ; D_SPcd[2] = -cte_e  ;
    D_SPcg[0] = -cte_f  ; D_SPcg[1] = 0.    ; D_SPcg[2] = -cte_e  ;
    D_SPrd[0] =  cte_f  ; D_SPrd[1] = 0.    ; D_SPrd[2] = -cte_e  ;
    D_SPrg[0] = -cte_f  ; D_SPrg[1] = 0.    ; D_SPrg[2] = -cte_e  ;
    D_PGad[0] =  cte_g  ; D_PGad[1] = 0.    ; D_PGad[2] = cte_h   ;
    D_PGag[0] = -cte_g  ; D_PGag[1] = 0.    ; D_PGag[2] = cte_h   ;
    D_PGcd[0] =  cte_g  ; D_PGcd[1] = 0.    ; D_PGcd[2] = cte_h   ;
    D_PGcg[0] = -cte_g  ; D_PGcg[1] = 0.    ; D_PGcg[2] = cte_h   ;
    D_PGrd[0] =  cte_g  ; D_PGrd[1] = 0.    ; D_PGrd[2] = cte_h   ;
    D_PGrg[0] = -cte_g  ; D_PGrg[1] = 0.    ; D_PGrg[2] = cte_h   ;
    GRd[0] =  cte_s     ; GRd[1] = 0.       ; GRd[2] = cte_t      ;
    GRg[0] = -cte_s     ; GRg[1] = 0.       ; GRg[2] = cte_t      ;
    nulvec (D_TS) ;

    conf_rov[0] =  47.7 D2RAD ; /* alfa_a */
    conf_rov[2] = -120. D2RAD ; /* beta_a */
    conf_rov[4] =  72.3 D2RAD ; /* gamma_a */
    conf_rov[1] = -47.7 D2RAD ; /* alfa_r */
    conf_rov[3] =  120. D2RAD ; /* beta_r */
    conf_rov[5] = -72.3 D2RAD ; /* gamma_r */
	for (i = 6 ; i <= 14; i++) conf_rov[i] =  0. D2RAD ; /* 2 theta puis 6 delta */

    xyzltr[0] = 0. ;
    xyzltr[1] = 0. ;
    xyzltr[2] = 0.7 ;
    xyzltr[3] = 0. D2RAD ;
    xyzltr[4] = 0. D2RAD ;
    xyzltr[5] = 0. D2RAD ;

	/* Calcul POR des diffrents corps */
    geodir_rover (xyzltr, conf_rov) ;
}


/*--------------------------------------------------------------------------*/
/*                         POUR LE GRAPHIQUE                                */
/*--------------------------------------------------------------------------*/
#define cte_lp 0.05
#define cte_hp 0.3
#define cte_hl 0.25
static int FirstTRa, FirstTRr, FirstTRm, FirstLIa,FirstLIr;

int makpt_roued (int n, double *porg, double *or, double phi)
{
int ii ;
double a, sa, t[3], m[3], axe[3] ;
void stkpoint (int, double, double, double) ;

    madvec (axe, porg+3, URd) ;
    scavec (axe, cte_rsp, axe) ;

    for (ii = 0 ; ii < 16 ; ii++)
    {
	a = phi + ii*M_PI/8.;
	sa = sin(a) ;
	t[0] = sa * sin_car ;
	t[1] = cos(a) ;
	t[2] = -sa * cos_car ;
	madvec (t, porg+3,t) ;
	cumvec (m, or, cte_rho, t) ;
	stkpoint (n++, m[0], m[1], m[2]) ; /* 23 a 38 */
	cumvec (t, or, cte_rcp, t) ;
	addvec (m, t, axe) ;
	stkpoint (n++, m[0], m[1], m[2]) ;
	subvec (m, t, axe) ;
	stkpoint (n++, m[0], m[1], m[2]) ;
    }
    return n ;
}

int makpt_roueg (int n, double *porg, double *or, double phi)
{
int ii ;
double a, sa, t[3], m[3], axe[3] ;
void stkpoint (int, double, double, double) ;

    madvec (axe, porg+3, URg) ;
    scavec (axe, cte_rsp, axe) ;

    for (ii = 0 ; ii < 16 ; ii++)
    {
	a = phi + ii*M_PI/8.;
	sa = sin(a) ;
	t[0] = -sa * sin_car ;
	t[1] = cos (a) ;
	t[2] = -sa * cos_car ;
	madvec (t, porg+3,t) ;
	cumvec (m, or, cte_rho, t) ;
	stkpoint (n++, m[0], m[1], m[2]) ; /* 23 a 38 */
	cumvec (t, or, cte_rcp, t) ;
	addvec (m, t, axe) ;
	stkpoint (n++, m[0], m[1], m[2]) ;
	subvec (m, t, axe) ;
	stkpoint (n++, m[0], m[1], m[2]) ;
    }
    return n ;
}

int makpt_bras(int n, double *jtronc, double *porp, double *oq, double *porg,
						      double *ug, double *or)
{
double t[3], m[3], j[3] ;
void stkpoint (int, double, double, double) ;

    cpyvec (j, jtronc) ;
    cpyvec (t, porp) ;

    cumvec (m, t, -cte_lp, j) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 16 */
    cumvec (m, t, cte_lp, j) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 17 */

    cpyvec (j, porp+6) ;
    cpyvec (t, oq) ;

    cumvec (m, t, -cte_lp, j) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 18 */
    cumvec (m, t, cte_lp, j) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 19 */

    cpyvec (m, porg) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 20 */
    madvec (t, porp+3, ug) ;
    cumvec (m, m, cte_m, t) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 21 */

    cpyvec (t, or) ;
    stkpoint (n++, t[0], t[1], t[2]) ; /* 22 */
    return n ;
}

int makpt_tronc(int n, double *post, double *orit)
{
double t[3], m[3], i[3], j[3], k[3] ;

void stkpoint (int, double, double, double) ;

    cpyvec (i, orit) ; cpyvec (j, orit+3) ; cpyvec (k, orit+6) ;
    cpyvec (t, post) ;
    cumvec (m, t, -cte_a, j) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 0 */
    cumvec (m, t, -cte_lp, j) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 1 */
    cumvec (m, t, cte_lp, j) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 2 */
    cumvec (m, t, cte_a, j) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 3 */

    cumvec (t, t, cte_xh, i) ;
    cumvec (t, t, cte_zh, k) ;
    cumvec (m, t, -cte_a, j) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 4 */
    cumvec (m, t, -cte_lp, j) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 5 */
    cumvec (m, t, cte_lp, j) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 6 */
    cumvec (m, t, cte_a, j) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 7 */

    cumvec (t, t, -(cte_xh+cte_xh), i) ;
    cumvec (m, t, -cte_a, j) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 8 */
    cumvec (m, t, -cte_lp, j) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 9 */
    cumvec (m, t, cte_lp, j) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 10 */
    cumvec (m, t, cte_a, j) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 11 */

    cumvec (t, t, cte_hp+cte_d-cte_zh, k) ;
    cumvec (m, t, -cte_a, j) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 12 */
    cumvec (m, t, cte_a, j) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 13 */

    cumvec (t, t, cte_xh+cte_xh, i) ;
    cumvec (m, t, -cte_a, j) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 14 */
    cumvec (m, t, cte_a, j) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 15 */

    return n ;
}

int makpt_liaison (int n, double *i, double *k,
					double *oa, double *ob, double *oc)
{
double m[3] ;
void stkpoint (int, double, double, double) ;

    cumvec (m, oa, -cte_lp, i) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 186 + 0 */
    cumvec (m, m, cte_hl, k) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 1 */
    cumvec (m, oa, cte_lp, i) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 2 */
    cumvec (m, m, cte_hl, k) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 3 */

    cumvec (m, ob, -cte_lp, i) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 4 */
    cumvec (m, m, cte_hl, k) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 5 */
    cumvec (m, ob, cte_lp, i) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 6 */
    cumvec (m, m, cte_hl, k) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 7 */

    cumvec (m, oc, -cte_lp, i) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 8 */
    cumvec (m, oc, cte_lp, i) ;
    stkpoint (n++, m[0], m[1], m[2]) ; /* 9 + 186  = 195 */
    return n ;
}

int makpointsRov (int N)
{
int	n = N;
/* tronc avant */
	FirstTRa = N;
    n = makpt_tronc (n, D_OSa, D_OCa+3) ;
    n = makpt_bras (n, D_OSa+6, D_OPad, OQad, D_OGad, UGd, ORad) ;
    n = makpt_roued (n, D_OGad, ORad, conf_rov[15]) ;
    n = makpt_bras (n, D_OSa+6, D_OPag, OQag, D_OGag, UGg, ORag) ;
    n = makpt_roueg (n, D_OGag, ORag, conf_rov[16]) ;

/* tronc arriere */
	FirstTRr = n;
    n = makpt_tronc (n, D_OSr, D_OCr+3) ;
    n = makpt_bras (n, D_OSr+6, D_OPrd, OQrd, D_OGrd, UGd, ORrd) ;
    n = makpt_roued (n, D_OGrd, ORrd, conf_rov[19]) ;
    n = makpt_bras (n, D_OSr+6, D_OPrg, OQrg, D_OGrg, UGg, ORrg) ;
    n = makpt_roueg (n, D_OGrg, ORrg, conf_rov[20]) ;

/* tronc central */
	FirstTRm = n;
    n = makpt_tronc (n, D_OT, D_OT+3) ;
    n = makpt_bras (n, D_OSc+6, D_OPcd, OQcd, D_OGcd, UGd, ORcd) ;
    n = makpt_roued (n, D_OGcd, ORcd, conf_rov[17]) ;
    n = makpt_bras (n, D_OSc+6, D_OPcg, OQcg, D_OGcg, UGg, ORcg) ;
    n = makpt_roueg (n, D_OGcg, ORcg, conf_rov[18]) ;

/* Liaison avant */
	FirstLIa = n;
    n = makpt_liaison (n, D_OT+3, D_OT+9, D_OAa, D_OBa, D_OCa) ;

/* Liaison arriere */
	FirstLIr = n;
    n = makpt_liaison (n, D_OT+3, D_OT+9, D_OAr, D_OBr, D_OCr) ;
    return n ;
}

void makeroue (int k)
{
int ii, k0 ;
void makeline(int,int) ;
void makequadri (int,int,int,int,int) ;
void maketrig (int,int,int,int) ;

    makeline (k-1, k) ;
    makeline (k-1, k+12) ;
    makeline (k-1, k+24) ;
    makeline (k-1, k+36) ;
    k0 = k ;
    for (ii = 0 ; ii < 15; ii++, k += 3)
    {
	maketrig (k, -(k+5), k+2, 1) ;
	maketrig (k+3, -(k+5), -k, 1) ;
	maketrig (k+4, -k, k+1, 1) ;
	maketrig (k, -(k+4), -(k+3), 1) ;

    }
    maketrig (k, -(k0+2), k+2, 1) ;
    maketrig (k0, -(k0+2), -k, 1) ;
    maketrig (k0+1, -k, k+1, 1) ;
    maketrig (k, -(k0+1), -k0, 1) ;
}

void maketronc (int k)
{
int j ;
void makeline(int,int) ;
void makequadri (int,int,int,int,int) ;
void maketrig (int,int,int,int) ;

    makequadri(k+0,k+3,k+7,k+4,0) ; /* cote bas droit */
    makequadri(k+0,k+8,k+11,k+3,0); /* cote bas gauche */

    maketrig(k+4,-(k+8),k+0,0) ;    /* cote arriere */
    maketrig(k+12,-(k+4),k+14,0) ;
    maketrig(k+8,-(k+4),-(k+12),0) ;

    maketrig(k+11,-(k+7),k+3,0) ;      /* cote avant */
    maketrig(k+7,-(k+11),-(k+15),0) ;
    maketrig(k+15,-(k+11),k+13,0) ;

    makequadri(k+4,k+7,k+15,k+14,0) ; /* cote haut droit */
    makequadri(k+11,k+8,k+12,k+13,0) ; /* cote haut gauche */
    makequadri(k+14,k+15,k+13,k+12,0) ; /* cote superieur */

/* cote droit */
    j = k+16 ;
    makequadri(k+1,k+2,j+1,j,1) ; /* bras inferieur droit */
    makequadri(k+5,k+6,j+3,j+2,1) ; /* bras superieur droit */

    makequadri(j+1,j,j+2,j+3,0) ;
    makequadri(j,j+4,j+5,j+2,0) ; /* pivot */
    makequadri(j+4,j+1,j+3,j+5,0) ;
    maketrig(j+2,j+5,j+3,0) ;
    maketrig(j+4,j,j+1,0) ;
    makeline (j+4,j+6) ; /* moyeu */
    makeroue(j+7) ; /* 1er point peripherie */
/* cote gauche */
    j = k + 71 ;
    makequadri(k+1,k+2,j+1,j,1) ; /* bras inferieur gauche */
    makequadri(k+9,k+10,j+3,j+2,1) ; /* bras superieur gauche */

    makequadri(j+1,j+3,j+2,j,0) ;
    makequadri(j,j+2,j+5,j+4,0) ; /* pivot */
    makequadri(j+4,j+5,j+3,j+1,0) ;
    maketrig(j+2,j+3,j+5,0) ;
    maketrig(j+4,j+1,j,0) ;
    makeline (j+4,j+6) ; /* moyeu */
    makeroue (j+7) ;
}

void makeliaison (int k)
{
void makequadri (int,int,int,int,int) ;
void maketrig (int,int,int,int) ;
void makeline(int,int) ;

    makequadri(k,k+4,k+6,k+2,1);
    makequadri(k+1,k+5,k+7,k+3,1);
    makequadri(k+4,k+5,k+7,k+6,1);
    makequadri(k+4,k+6,k+9,k+8,1);
}

void makerover (void)
{
void makequadri (int,int,int,int,int) ;
void maketrig (int,int,int,int) ;
void makeline(int,int) ;

    maketronc (FirstTRa) ;
    maketronc(FirstTRr) ;
    maketronc (FirstTRm) ;
    makeliaison(FirstLIa) ;
    makeliaison(FirstLIr) ;
}
