/* SHOWROV : VISU DU ROVER AVEC LIGNES CACHEES */
/*

A l'initialisation :

  Aprs allocations initiales et calcul de constantes, on calcule :
  - les situations des diffrents corps du vhicules en fonction de sa configuration, 
  - les coordonnes/Ro de TOUS les points qui dcrivent le corps : "npt" points "POINTI",
  puis on dcrit :
  - les faces : nbtri "TRIGLE" et les artes : nbaret "ARETES",
  puis on calcule :
  - les coordonnes de TOUS ces points dans le repre camra :"npt" points "POINTP",

Dans la boucle d'animation :
  on calcule :
  - les lignes  tracer (qu'on envoie  la mmoire partage)
  on modifie ventuellement le point de vue ou la configuration et on calcule :
  - les situations des diffrents corps du vhicules en fonction de sa configuration, 
  - les coordonnes/Ro de TOUS les points qui dcrivent le corps : "npt" points "POINTI",
  - les coordonnes de TOUS ces points dans le repre camra :"npt" points "POINTP"
  - retour dbut de boucle.

Principe :

  Le tableau ARETES dcrit toutes les artes du corps au moyen de 3 entier :
  - les 2 numros des points extrmes de l'arte (le 1er est toujours < au 2eme)
  - un flag de visibilit potentielle :
	 0 l'arte n'est pas visible (sur la face arrire d'un corps)
	 1 l'arte est potentiellement visible (sur la face avant d'un corps)
	-1 l'arte est potentiellement visible (c'est un segment de ligne isol).

  Le tableau TRIGLE dcrit les triangles issus du dcoupage des faces des corps. Un
  triangle comporte :
  - les 3 numros de ses 3 sommets
  - l'quation normalise dans le repre camra du plan qui contient le triangle
    (vecteur unitaire orthogonal et distance signe du plan  la camra)
  - les numros de 3 artes qui constituent les 3 cots du triangle (-1 si l'arte
    n'existe pas, par exemple diagonale d'un rectangle dcompos en 2 triangles).
  - un flag mis  1 si la face est une plaque visible des deux cots (au contraire des
    faces d'un corps volumique qui ne sont visibles que du cot extrieur).
  - un flag mis  0 lorsque le triangle est la premier d'un groupe de triangles conscutifs
    dcrivants un polygone. L'quation du plan du polygone est calcule  partir des 3
	points de ce triangle et est conserve pour tous les autres de ce polygone.

A l'initialisation, on cre :
	- les artes isoles (visiblilt  -1)
	- les triangles et les artes priphriques (visibilit  0)

Dans la boucle de visualisation :
	- on calcule les positions des points,
	- on divise l'cran en zones rectangulaires et on cre des listes (vides) des triangles qui
	  recouvrent ces zones
	- on calcule les quations des triangles, on met  1 la visibilit des artes des faces doubles
	  ou des faces avant, on affecte les triangles aux listes des zones concernes.
	Puis, pour chaque arte dont la visilit est  1 on examine (zone par zone) si des triangles ne
	la cachent pas en tout ou partie, pour gnrer des ordres de trac correspondants aux
	parties qui sont visibles.
*/

#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <float.h>
#include <ctype.h>
#include <stdlib.h>
#include <math.h>
#include "geouti.h"
#include "showrov.h"

#ifndef M_PI
#define	M_PI  3.14159265358979323846
#endif

#ifndef max
#define max(x,y) ((x) > (y) ? (x) :(y))
#define min(x,y) ((x) < (y) ? (x) :(y))
#endif
#define max3(x,y,z) ((x) > (y) ? max(x,z) : max(y,z))
#define min3(x,y,z) ((x) < (y) ? min(x,z) : min(y,z))
#define xwhole(x) ((int) (((x)-Xmin)/deltaX))
#define ywhole(y) ((int) (((y)-Ymin)/deltaY))
#define xreal(i) (Xmin+(i)*deltaX)

#define D2RAD *M_PI/180.

#define GLOBAL
GLOBAL double conf_cam[10] ;
extern double conf_rov[21] ;
extern double xyzltr[6];
static double matrot[9];

typedef struct { double x,y,z ;          /* Coordonnees des points */
               } type_point ;

typedef struct {
		 int A,B,C ;        /* Numero des trois sommets du triangle */
		 double a,b,c,h ;/* Equation normalisee du plan du triangle */
		 int naret[3];    /* numero des trois aretes (<0 -> neant) */
		 int fils;       /* 0 si pere et autre si fils (appartient au plan d'un pere) */
		 int dble;       /* Non nul si double face (toujours un cot visible)*/
	       } type_trigle ;                /* ax + by + cz = h dans Rcam */

typedef struct node { int jtr;        /* N d'un triangle couvrant une case */
                      struct node *next ;           /* pointeur sur suivant */
               } type_node ;

typedef struct { int tr_cov ;/* N du triangle le + prs recouvrant la case */
                 double tr_dist ;    /* Distance de ce triangle  la camera */
                 type_node *start ; /* Liste des triangles couvrant la case */
               } type_case ;

typedef struct { 
	int ori;
    int des ;
	int visible; /* -1 tjs visible, 0 cache, 1 visible */
               } type_arete ;

/*-------------------------------------------------------------------------------------
   Pour simplifier la rgle de trois de la projection sur l'cran, on prend Zecran = 1.
   pour  distance camra-cran :
   => Xecran = Zecran*Xuser/Zuser , se simplifie en Xecran = Xuser/Zuser
   La scene 2D construite sur l'ecran est limite   une taille TGEVUE x TGEVUE
   La valeur TGEVUE = 1 correspond  un angle de vision de 45 degrs. 
   On met l'origine au centre de l'cran.
---------------------------------------------------------------------------------------*/

#define NCAR 40					/* Nb de carres par ligne ou colonne de l'ecran */
#define TGEVUE 1.
GLOBAL double TgVue = TGEVUE;
static double Xmin = -TGEVUE/2.;
static double Ymin = -TGEVUE/2.;
static double deltaX = TGEVUE/NCAR;
static double deltaY = TGEVUE/NCAR;

static double yMinfini = -1.e6 ;/* Point tres loin pour test devant/derriere */
static double eps=1.e-5, meps = -1.e-5, unpeps = 1.+1.e-5;

static int nbtri, nbtri1, nbtri2, nbaret, nbaret1, nbaret2, npt_tot1, npt_tot2,
    nbset, nbsetmax = 0,
    ipixmin, ipixmax, ipixleft, ipixright, ipix,
    jpix, jtop, jbot, j_old, I, jI ;
static int LOWER[NCAR], UPPER[NCAR], LOW[NCAR], UP[NCAR] ;
/* Pour mmoriser les num des sommets avant d'appeler polymak */
#define MAXSPOLMK 3 /* Nombre max de sommet d'un polygone triangularis avant d'appeler polymak */
static int Sommet[MAXSPOLMK]; 

#define BYALLOC
/* Pour dimensionner, faire des tests en dfinissant BYALLOC. A la fin de l'execution
   noter les tailles maximales utilisees, les majorer de 10 a 20 % pour definir les
   dimensions ci-dessous, et commenter la dfinition de BYALLOC */
/* Taille initiale des tableaux allous */
#define SIZEINI		200 
#define MAXPOINTS	600
#define MAXTRIGLES	800
#define MAXTRSET	600
#define MAXARETES	1200
#define MAXNTPLOT	1200

#ifdef BYALLOC

static int nbptsup = SIZEINI,
		   nbtrisup = SIZEINI,
		   nbsetsup = SIZEINI,
		   nbaretsup = SIZEINI;
static type_point *POINTP = NULL ; /* Doit tre initialise a NULL !!!*/
type_point *POINTI ;
static type_trigle *TRIGLE ;
static type_arete *ARETES ;
static int *TRSET;

#else

static int nbptsup = MAXPOINTS,
		   nbtrisup = MAXTRIGLES,
		   nbsetsup = MAXTRSET,
		   nbaretsup = MAXARETES;
static type_point POINTP[MAXPOINTS] ;
type_point POINTI[MAXPOINTS] ;
static type_trigle TRIGLE[MAXTRIGLES] ;
static type_arete ARETES[MAXARETES] ;
static int TRSET[MAXTRSET];

#endif

/* Pour basculer de buffer de trac */
extern double sizemeta, szms2;
extern HENHMETAFILE	hemf;
extern RECT RectMeta ;
GLOBAL HDC hdc_cou;

static type_point *ppoint ;
static type_trigle *ptrigle ;
static type_node *pnode ;
static type_case CASES[NCAR*NCAR], *ppix ;

void UpdateDraw(void);
void MyMessBox(char *);
/*************************************************************************************/
/* SORTIE AVEC MESSAGE ERREUR */
static void exitmes (char *str) { MyMessBox(str) ; exit (1) ; }


/*************************************************************************************/
/* Ajout du segment PQ dans le tableau des artes */
static int add_arete (int P, int Q, int vis)
{
int i, iaux;
type_arete *paret;
	
    if (P > Q) {iaux = P ; P = Q ; Q = iaux ; }
	/* On renvoi son numero si elle existe deja */
	for (i = 0 ; i < nbaret; i++)
		if (ARETES[i].ori == P && ARETES[i].des == Q) return i;

	/* Elle n'existe pas, on la cree */
#ifdef BYALLOC
    if (nbaret >= nbaretsup)
    {
		nbaretsup *= 2 ;
		ARETES = (type_arete *) realloc (ARETES, nbaretsup*sizeof(type_arete)) ;
		if (ARETES == NULL) exitmes ("Echec realloc ARETES") ;
    }
#endif


	paret = &ARETES[nbaret++];
	paret->ori = P;
	paret->des = Q;
	if (vis) paret->visible = -1;
	else paret->visible = 0;
	return i;
}

static void add_trig (int i0, int i1, int i2, int dble, int fils)
{
int A,B,C;
    A = abs (Sommet[i0]);
    B = abs (Sommet[i1]);
    C = abs (Sommet[i2]);

#ifdef BYALLOC
    if (nbtri == nbtrisup)
    {
		nbtrisup *= 2 ;
		TRIGLE = (type_trigle *) realloc (TRIGLE, nbtrisup*sizeof(type_trigle)) ;
		if (TRIGLE == NULL) exitmes ("Echec realloc TRIGLE") ;
    }
#endif
    ptrigle = &TRIGLE[nbtri++] ;
	ptrigle->fils = fils;
	ptrigle->dble = dble;
    ptrigle->A = A ; ptrigle->B = B ; ptrigle->C = C ;
	ptrigle->naret[0] = -1;
	ptrigle->naret[1] = -1;
	ptrigle->naret[2] = -1;
}

/*************************************************************************************/
double orient (type_point *pA, type_point *pB, type_point *pC)
{

    return (pA->x * (pB->y * pC->z - pC->y * pB->z) -
			pB->x * (pA->y * pC->z - pC->y * pA->z) +
			pC->x * (pA->y * pB->z - pB->y * pA->z)) ;
}


static int caldiag (int i0, int i1, int i2, double h0, double *diag)
{
/* Calcule la diagogonale */
double xdist, ydist, zdist;
type_point *pA, *pB, *pC;

	*diag = DBL_MAX ;

    pA = &POINTI[abs (Sommet[i0])];
    pB = &POINTI[abs (Sommet[i1])];
    pC = &POINTI[abs (Sommet[i2])];

	/* Retour si point interieur */
	if (h0*orient(pA, pB, pC) < 0) return 0; /* Point intrieur, triangle  rejeter */

	xdist = pC->x - pA->x ; 
	ydist = pC->y - pA->y ; 
	zdist = pC->z - pA->z ;
	*diag = xdist*xdist + ydist*ydist + zdist*zdist ;
	return 1; /* bon triangle */
}

/*************************************************************************************
   Traitement d'un triangle : On calcule :
   1. h0 : La distance de la cam au 1er triangle d'un plan
   2. On active si artes si visible
   3. On calcul l'quation du plan du 1er triangle du plan
   4. On met  jour le tableau CASE qui recense les triangles recouvrant les carrs
***************************************************************************************/
static void exam_trig (int ntri)
{
int i, k, A,B,C, topcode[3];

double denom, pente, Xleft[3], Xright[3], Yleft[3], Yright[3],
    xA, yA, zA, xB, yB, zB, xC, yC, zC, r,
    XA, YA, XB, YB, XC, YC, DA, DB, DC, D, DAB, DAC, DBC,
    aux, dist, xR, yR, h0 ;
static double a, b, c, h ;
static int visible; /* On conserve la valeur du pere calcule en premier */

	ptrigle = &TRIGLE[ntri] ;
	if (!ptrigle->fils) /* C'est un pre : On calcule sa position / Camera */
	{ 
		A = ptrigle->A ; B = ptrigle->B  ; C = ptrigle->C ;
		ppoint = &POINTP[A] ;
		xA = ppoint->x , yA = ppoint->y ; zA = ppoint->z ;
		ppoint = &POINTP[B] ;
		xB = ppoint->x , yB = ppoint->y ; zB = ppoint->z ;
		ppoint = &POINTP[C] ;
		xC = ppoint->x , yC = ppoint->y ; zC = ppoint->z ;

		/* distance camera au plan du triangle */
		h0 = xA * (yB * zC - yC * zB) -	 xB * (yA * zC - yC * zA) +
										 xC * (yA * zB - yB * zA) ;

		if (ptrigle->dble)
		{
			if (fabs(h0) < eps) visible = 0; /* plaque passant par la camera */
			else visible = 1;
		}
		else 
		{
			if (h0 < eps) visible = 0; /* triangle arrire */
			else visible = 1;
		}
		if (visible) /* On calcule sa normale */
		{
			a =   yA * (zB - zC) - yB * (zA - zC) + yC * (zA - zB) ;
			b = -(xA * (zB - zC) - xB * (zA - zC) + xC * (zA - zB)) ;
			c =   xA * (yB - yC) - xB * (yA - yC) + xC * (yA - yB) ;
			r = sqrt (a*a+b*b+c*c) ; if (r == 0.) r = eps ;
			h = h0/r ; a /= r ; b /= r ; c /= r ;
			ptrigle->a = a ; ptrigle->b = b ; ptrigle->c = c ; ptrigle->h = h ;
		}
	}
	if (!visible) return ; /* derriere la camera */


	if (ptrigle->fils) /* C'est un fils. On recupre ses sommets */
	{
		A = ptrigle->A ; B = ptrigle->B  ; C = ptrigle->C ;
		ppoint = &POINTP[A] ;
		xA = ppoint->x , yA = ppoint->y ; zA = ppoint->z ;
		ppoint = &POINTP[B] ;
		xB = ppoint->x , yB = ppoint->y ; zB = ppoint->z ;
		ppoint = &POINTP[C] ;
		xC = ppoint->x , yC = ppoint->y ; zC = ppoint->z ;
		/* meme normale que le pere */
		ptrigle->a = a ; ptrigle->b = b ; ptrigle->c = c ; ptrigle->h = h ;
	}

	/* On active ses artes visibles */
	for (i=0;i<3;i++)
	{
		k = ptrigle->naret[i];
		if (k < 0) break;
		ARETES[k].visible = 1;
	}

    /* memo du triangle dans la liste ecran */
    XA = xA/zA ; YA = yA/zA ;
    XB = xB/zB ; YB = yB/zB ;
    XC = xC/zC ; YC = yC/zC ;
    DA = XB*YC-XC*YB ;DB = XC*YA - XA*YC ; DC = XA*YB-XB*YA ; D = DA+DB+DC ;
    DAB = DC-yMinfini*(XA-XB) ; DAC = DB-yMinfini*(XC-XA) ; DBC = DA-yMinfini*(XB-XC) ;
	/* topcode = 1 si cot suprieur et 0 si cot infrieur (/  haut-bas de l'cran) */
    topcode[0]=(D*DAB>0) ;
    topcode[1]=(D*DAC>0) ;
    topcode[2]=(D*DBC>0) ;
    Xleft[0]=XA ; Yleft[0]=YA ; Xright[0]=XB ; Yright[0]=YB ;
    Xleft[1]=XA ; Yleft[1]=YA ; Xright[1]=XC ; Yright[1]=YC ;
    Xleft[2]=XB ; Yleft[2]=YB ; Xright[2]=XC ; Yright[2]=YC ;
    for (I=0 ; I<3 ; I++) /* I = numero de cote du triangle */
    {
		if ( Xleft[I] > Xright[I] ||
			 (Xleft[I] == Xright[I] && Yleft[I] > Yright[I]))
		{
			aux = Xleft[I] ; Xleft[I] = Xright[I] ; Xright[I] = aux ;
			aux = Yleft[I] ; Yleft[I] = Yright[I] ; Yright[I] = aux ;
		}
    }
	/* Index des cases potentiellement concernes par ABC */
    ipixmin = xwhole (min3(XA, XB, XC)) ;
    ipixmax = xwhole (max3(XA, XB, XC)) ;
    for (ipix = ipixmin ; ipix <= ipixmax ; ipix++)
    {
		if (ipix < 0 || ipix >= NCAR) continue;
		LOWER[ipix] = UP[ipix]  =  NCAR ;
		UPPER[ipix] = LOW[ipix] = -1 ;
    }
	/* 
	   Examen des 3 cots AB, BC et CA pour construire LOWER,UPPER,LOW et UP :
	   LOWER[ipix] : range la plus basse touche par le bas  du triangle dans la colonne ipix
	   UPPER[ipix] : range la plus haute touche par le haut du triangle dans la colonne ipix

       UP[ipix]    : range la plus basse touche par le haut du triangle dans la colonne ipix
       LOW[ipix]   : range la plus haute touche par le bas  du triangle dans la colonne ipix

       Dans la colonne ipix, le triangle est inscrit dans les ranges LOWER  UPPER
	   et le triangle inscrit les ranges LOW  UP.
	*/
    for (I = 0 ;I < 3 ; I++)
    {
		ipixleft = xwhole (Xleft[I]) ; ipixright = xwhole (Xright[I]) ;
		denom = Xright[I]-Xleft[I] ;
		if (ipixleft != ipixright) pente = (Yright[I]-Yleft[I])/denom ;
		j_old = ywhole (Yleft[I]) ;
		for (ipix = ipixleft ; ipix <= ipixright ; ipix++)
		{
			if (ipix < 0 || ipix >= NCAR) continue;
			/* jI : Numro case verticale  l'intersection avec la paroi gauche de la case
			        sauf pour la dernire, ou jI est le numro au niveau du point droit */
			if (ipix == ipixright) jI = ywhole (Yright[I]) ;
			else jI = ywhole (Yleft[I] + pente * (xreal(ipix+1) -Xleft[I])) ;
			if (topcode[I])
			{
				UPPER[ipix] = max3 (j_old, jI, UPPER[ipix]) ;
				UP[ipix] = min3 (j_old, jI, UP[ipix]) ;
			}
			else
			{
				LOWER[ipix] = min3 (j_old, jI, LOWER[ipix]) ;
				LOW[ipix] = max3 (j_old, jI, LOW[ipix]) ;
			}
			j_old = jI ;
		}
    }

    /* Pour la colonne ecran ipix, le triangle est associe seulement avec
       les carres de la ligne LOWER[ipix], ... , UPPER[ipix].
       Le sous-ensemble LOW[ipix]+1, ... , UP[ipix]-1 de ces lignes concerne
       des carres qui appartiennent completement au triangle */
    for (ipix = ipixmin ; ipix <= ipixmax ; ipix++)
	{
		if (ipix < 0 || ipix >= NCAR) continue;
		for (jpix = LOWER[ipix] ; jpix <= UPPER[ipix] ; jpix++)
		{
			if (jpix < 0 || jpix >= NCAR) continue;
			ppix = &CASES[ipix*NCAR+jpix] ;
			if (jpix > LOW[ipix] && jpix < UP[ipix])
			{ /* Pour un carr compltement inscrit dans le triangle 
				 distance au centre du carr */
				xR = Xmin+(ipix+0.5)*deltaX ;
				yR = Ymin+(jpix+0.5)*deltaY ;
				denom = a*xR + b*yR + c ;
				dist = fabs (denom) > eps
				   ? h * sqrt (xR*xR + yR*yR + 1.) / denom : DBL_MAX ;
			/* La ligne de l'oeil E au carre (xR, yR, 1) coupe le plan ABC
			   a la distance dist de E */
				if (dist < ppix->tr_dist)
				{ ppix->tr_cov = ntri ; ppix->tr_dist = dist ; }
			}
			else /* Ajoute le triangle a la liste ecran */
			{
				pnode = (type_node *) malloc (sizeof (type_node)) ;
				if (pnode == NULL) exitmes ("Echec allocation pnode #1") ;
				pnode->jtr = ntri ; pnode->next = ppix->start ;
				ppix->start = pnode ;
			}
		}
	}
}

/*************************************************************************************/
/* Examen visibilite segment PQ /aux triangles TRSET[jO] a TRSET[nbset-1] */
static void visibilite (double xP, double yP, double zP,
		 double xQ, double yQ, double zQ, int k0)
{
int j, k  = k0, afaire = 1, A, B, C, i,
    Phors, Qhors, out, Pout, Qout, eA, eB, eC, som ;
double a, b, c, h, hP, hQ, r1, r2, r3,
    xA, yA, zA, xB, yB, zB, xC, yC, zC, dA, dB, dC,
    Lmin, Lmax, Lab, Mu, xmin, ymin, zmin, xmax, ymax, zmax,
    C1, C2, C3, K1, K2, K3, denom1, denom2, Cpos, Ppos, Qpos, aux, eps1 ;

double gain;
int x, y;

	/* k est l'index du trig qui va tre compar  PQ */
    while (k < nbset) /* Tant que k est infrieur au nb de trig susceptibles de cacher PQ */
    {
		/* Recupere les caractristiques du triangle */
		j = TRSET[k] ; ptrigle = &TRIGLE[j] ;
		a = ptrigle->a ; b = ptrigle->b ; c = ptrigle->c ;
		h = ptrigle->h ;
		if (h < 0.) { a = -a ; b = -b ; c = -c ; h = -h ; }
		eps1 = eps + eps * h ;
		/* Test 1 : P et Q sont-ils tous deux tous deux devant le plan ABC */
		hP = a * xP + b * yP + c * zP ; hQ = a * xQ + b * yQ + c * zQ ;

		if (hP <= h+ eps1 && hQ <= h + eps1)
		{ /* Oui : examen des autres plans */ k++ ; continue ; }

		/* Test 2 : La droite PQ est elle exterieure a la pyramide E_ABC */
		/* Pour cela : test des points A, B et C par rapport au plan EPQ */
		/* Equation du plan EPQ : K1 x + K2 y + K3 z = 0 */
		K1 = yP*zQ - yQ*zP ; K2 = zP*xQ - zQ*xP ;  K3 = xP*yQ - xQ*yP ;
		A = ptrigle->A ; B = ptrigle->B ; C = ptrigle->C ;
		ppoint = &POINTP[A] ;
		xA = ppoint->x ; yA = ppoint->y ; zA = ppoint->z ;
		ppoint = &POINTP[B] ;
		xB = ppoint->x ; yB = ppoint->y ; zB = ppoint->z ;
		ppoint = &POINTP[C] ;
        xC = ppoint->x ; yC = ppoint->y ; zC = ppoint->z ;
        dA = K1*xA + K2*yA + K3*zA ;
        dB = K1*xB + K2*yB + K3*zB ;
        dC = K1*xC + K2*yC + K3*zC ;
        /* Si dA, dB, dC de meme signe, les points A, B, C sont du meme
           cote du plan EPQ */
        eA = dA > eps ? 1 : dA < meps ? -1 : 0 ;
        eB = dB > eps ? 1 : dB < meps ? -1 : 0 ;
        eC = dC > eps ? 1 : dC < meps ? -1 : 0 ;
        som = eA + eB + eC ;
        /* si |som| = 3 : PQ exterieur.
           si |som| = 2 : idem avec un des trois points A,B ou C qui touche
			  le plan EPQ. */
        if (abs (som) >= 2)
        { /* Oui : examen des autres plans */ k++ ; continue ; }

		/* Test 3 : Examen intersection de PQ avec plan EAB, EBC, ECA */
		Pout = Qout = 0 ; Lmin = 1. ; Lmax = 0. ;
		/* Test si P et Q cote oppose a C par rapport a EAB */
        for (i = 0 ; i < 3 ; i++) /* Test sur E_AB puis E_BC puis E_CA */
        {
            /* Equation du plan EAB : C1 x + C2 y + C3 z = 0 */
            C1 = yA*zB - yB*zA ; C2 = zA*xB - zB*xA ; C3 = xA*yB - xB*yA ;
            Cpos = C1*xC + C2*yC + C3*zC ;
            Ppos = C1*xP + C2*yP + C3*zP ;
            Qpos = C1*xQ + C2*yQ + C3*zQ ;
            if (Cpos > eps)
            {
                Phors = Ppos < meps ; Qhors = Qpos < meps ;
                out = Phors && Qpos <= eps || Qhors && Ppos <= eps ;
            }
            else if (Cpos < meps)
			{
                Phors = Ppos > eps ; Qhors = Qpos > eps ;
                out = Phors && Qpos >= meps || Qhors && Ppos >= meps ;
            }
            else out = 1 ;
			if (out) /* Oui examen autre plan */ break ;

            /* Intersection PQ avec EAB : M = P + Lab PQ */
            denom1 = Qpos - Ppos ;
            Lab = fabs (denom1) <= eps ? 1.e7 : -Ppos/denom1 ;
            /* Intersection AB avec EPQ : M = A + Mu AB */
            denom2 = dB - dA ;
            Mu = fabs (denom2) <= eps ? 1.e7 : -dA/denom2 ;

            /*  Si intersection segment : memo Lab min et max */
            if (Mu>=meps &&  Mu<=unpeps && Lab>=meps && Lab<=unpeps)
            { if (Lab < Lmin) Lmin = Lab ; if (Lab > Lmax) Lmax = Lab ; }

            /* Echange circulaire des point A, B, C */
            aux = xA ; xA = xB ; xB = xC ; xC = aux ;
			aux = yA ; yA = yB ; yB = yC ; yC = aux ;
            aux = zA ; zA = zB ; zB = zC ; zC = aux ;
            aux = dA ; dA = dB ; dB = dC ; dC = aux ;
            Pout |= Phors ;
            Qout |= Qhors ;
		}
		if (out) {/* PQ exterieur a E_ABC : plans suivants */ k++ ; continue ; }

		/* Test 4 : si P et Q completement interieurs a la pyramide E_ABC */
        if (!(Pout || Qout)) { /* PQ invisible */ afaire = 0 ; break ; }

		/* Test 5 : Le point d'intersection est-il devant le plan ABC */
		/* Un seul point teste car les aretes des solides sont supposees
		   exterieures aux autres solides (pas d'intersection de solides) */
		r1 = xQ-xP ; r2 = yQ-yP ; r3 = zQ-zP ;
        /* Coordonnees point intersection le + pres de P */
        xmin = xP+Lmin*r1 ; ymin = yP+Lmin*r2 ; zmin = zP+Lmin*r3 ;
        if (a*xmin+b*ymin+c*zmin<h-eps1)
		{ /* Oui devant : examen plans suivants */ k++; continue; }
        /* Coordonnees point intersection le + pres de Q */
		xmax = xP+Lmax*r1 ; ymax = yP+Lmax*r2 ; zmax = zP+Lmax*r3 ;
		if (a*xmax+b*ymax+c*zmax<h-eps1)
		{/* Oui devant : examen plans suivants */ k++; continue; }

		/* Test 6 : Examen des segments exterieurs a l'intersection si PI et/ou
            JQ sont exterieurs ou en avant de ABC */
        if (Pout || hP < h - eps1) visibilite (xP, yP, zP, xmin, ymin, zmin, k+1) ;
        if (Qout || hQ < h - eps1) visibilite (xQ, yQ, zQ, xmax, ymax, zmax, k+1) ;
        afaire = 0 ; break ;
    }

    if (afaire && (zP > 0. || zQ > 0.))
	{
		if (zP > 0.)
		{
			if (zQ > 0.)
			{
				xA = xP/zP;
				yA = yP/zP;
				xB = xQ/zQ;
				yB = yQ/zQ;
			}
			else /* zQ < 0 */
			{
				xA = xP/zP;
				yA = yP/zP;
				aux = (1. - zP)/(zQ - zP);
				xB = xQ + aux*(xQ - xP);
				yB = yQ + aux*(yQ - yP);
			}
			
		}
		else /* zP < 0 */
		{
			aux = (1. - zP)/(zQ - zP);
			xA = xQ + aux*(xQ - xP);
			yA = yQ + aux*(yQ - yP);
			xB = xQ/zQ;
			yB = yQ/zQ;
		}
		gain = sizemeta/TgVue ; /* associ au plus grand cot */
		x = (int) (szms2 + gain * xA);
		y = (int) (szms2 - gain * yA);
		MoveToEx(hdc_cou, (int) (szms2 + gain * xA), (int) (szms2 - gain * yA),  NULL); 
		x = (int) (szms2 + gain * xB);
		y = (int) (szms2 - gain * yB);
		LineTo (hdc_cou, (int) (szms2 + gain * xB), (int) (szms2 - gain * yB));
	}
}

/*************************************************************************************/
/* CHANGEMENT DE BASE OBJET -> VISION  */
static double *carol2mrot(double *a, double *carol)
{ // cap/Z, puis assiette/Y, puis roulis/x 
double cc, sc, ca, sa, cr, sr ;
double sasr, sacr ;

    cc = cos (carol[0]) ; ca = cos (carol[1]) ; cr = cos (carol[2]) ;
    sc = sin (carol[0]) ; sa = sin (carol[1]) ; sr = sin (carol[2]) ;
    sasr = sa * sr ; sacr = sa * cr ;

    a[0] = cc*ca  ; a[3] = cc*sasr - sc*cr ; a[6] = cc*sacr + sc*sr ;
    a[1] = sc*ca  ; a[4] = sc*sasr + cc*cr ; a[7] = sc*sacr - cc*sr ;
    a[2] =   -sa  ; a[5] =    ca*sr        ; a[8] =    ca*cr        ;
	return a;
}

static void obj2vue (double x, double y, double z, double *px, double *py, double *pz)
{
     x += conf_cam[3];
     y += conf_cam[4];
     z += conf_cam[5];

    *pz = conf_cam[8] - (matrot[0]*x + matrot[1]*y + matrot[2]*z);
    *px = conf_cam[6] +  matrot[3]*x + matrot[4]*y + matrot[5]*z ;
    *py = conf_cam[7] +  matrot[6]*x + matrot[7]*y + matrot[8]*z ;
}

/*************************************************************************************
 Triangularise un polygone et gnre les artes visibles (contour externe)
 npoly : nb de cots
 dblface : vrai si vu des 2 cts, sinon n'est opaque que du cot de la normale
           extrieure.
 Les numros des sommets sont dans le tableau Sommet[0  npoly-1]. Les sommets doivent 
 tre dcrits dans le sens positifs pour les contours externes et dans le sens ngatif
 pour les contours internes (les trous).
 Un numro ngatif correspond  l'extremit d'une arte fictive qui ne doit pas tre
 trace (arte interne qui n'appartient ni au contour externe, ni  un trou).
 Attention, si un sommet porte le numro 0, il ne peut pas tre  l'extrmit d'une
 arte invisible !!!!. Pour l'viter, on commence la numrotation  1.
 *************************************************************************************/
void polymak (int npoly, int dblface)
{
int i,j,k,n,ntri, code, pointnr, count, ii, i0, i1, i2, imin ;
int naret, P, Q;
double  diag, min_diag, h0 ;
type_point *pA, *pB, *pC;

/* Cas point isole : erreur */
    if (npoly <= 1) exitmes ("Point isole !") ;
/* Cas segment isole : toujours visible */
    if (npoly == 2) { add_arete (Sommet[0], Sommet[1], 1) ; return ; }

/* Cas triangle */
	k = 0;
	if (npoly == 3) 
	{ 
		ntri = nbtri; naret = nbaret;
		add_trig(0, 1, 2, dblface, 0) ;
		for (i = 1 ; i <= npoly ; i++)
		{
			ii = i % npoly ; code = Sommet[ii] ; pointnr = abs(code) ;
			if (code < 0) Sommet[ii] = pointnr ;
			else TRIGLE[ntri].naret[k++] = add_arete (Sommet[i-1], pointnr, 0) ;
		}
		return ;
	}

/* Cas vrai polygone */
	naret = nbaret;
	for (i = 1 ; i <= npoly ; i++)
	{
		ii = i % npoly ; code = Sommet[ii] ; pointnr = abs(code) ;
		if (code < 0) Sommet[ii] = pointnr ;
		else add_arete (Sommet[i-1], pointnr, 0) ;
	}
	/* Orientation donn par les 3 premier points */
    pA = &POINTI[abs (Sommet[0])],
    pB = &POINTI[abs (Sommet[1])],
    pC = &POINTI[abs (Sommet[2])],
	h0 = orient(pA, pB, pC) ;
    /* Division d'un polynome en triangles */
    count = 0 ;
    while (npoly > 2)
    {

		/* Recherche de la plus petite diagonale interne du polygone */
		min_diag = DBL_MAX ;
		for (i1 = 0 ; i1 < npoly ; i1++)
		{
			i0 = (i1 == 0 ? npoly-1 : i1 - 1) ;
			i2 = (i1 == npoly-1 ? 0 : i1 + 1) ;
			if (caldiag (i0, i1, i2, h0, &diag ) && diag < min_diag)
			{ min_diag = diag ; imin = i1 ; }
		}
		i1 = imin ;
		i0 = (i1 == 0 ? npoly-1 : i1 - 1) ;
		i2 = (i1 == npoly-1 ? 0 : i1 + 1) ;
		/* memo du triangle ds TRIGLE */
		ntri = nbtri;
		add_trig (i0, i1, i2, dblface, count++) ;
		/* Affectation des aretes aux triangles concerns */
		ii = 0;
		for (n=0;n<3;n++) /* Les 3 cots du triangle */
		{   /* On gnre les index des 2 extrmits */
			switch (n) 
			{
			case 0 : i = i0; j = i1; break;
			case 1 : i = i1; j = i2; break;
			case 2 : i = i2; j = i0; break;
			}
			/* On ordonne les 2 numros */
			if (Sommet[i] < Sommet[j]) {P = Sommet[i]; Q = Sommet[j];}
			else {P = Sommet[j]; Q = Sommet[i];}
			/* On cherche la correspondance avec les artes de ce polygone */
			for (k = 0 ; k < nbaret; k++)
				if (ARETES[k].ori == P && ARETES[k].des == Q) break;
			/* Si trouve on l'affecte au triangle */
			if (k < nbaret) TRIGLE[ntri].naret[ii++] = k;
		}
		/* On retire le sommet numro i1 de la liste des sommets */
		npoly-- ;
		for (ii = i1 ; ii <= npoly ; ii++) Sommet[ii] = Sommet[ii+1] ;
    }
}

  

/*************************************************************************************/
void makeline (int i, int j)
{
    Sommet[0] = i ; Sommet[1] = j ; polymak (2,0) ;
}

void maketrig (int i, int j, int k, int dbl)
{
    Sommet[0] = i ; Sommet[1] = j ; Sommet[2] = k; polymak (3,dbl) ;
}

void makequadri (int i, int j, int k, int l,  int dbl)
{
    Sommet[0] = i ; Sommet[1] = -k ; Sommet[2] = l ; polymak (3,dbl) ;
    Sommet[0] = k ; Sommet[1] = -i ; Sommet[2] = j ; polymak (3,dbl) ;
}

/*************************************************************************************/
void stkpoint (int k, double x, double y, double z)
{
#ifdef BYALLOC
    if (k >= nbptsup)
    {
	nbptsup = nbptsup*2 > k ? nbptsup*2 : k+1 ;
	POINTI = (type_point *) realloc (POINTI, nbptsup*sizeof(type_point)) ;
	if (POINTI == NULL) exitmes ("Echec realloc POINTI");
    }
#endif
    POINTI[k].x = x ; POINTI[k].y = y ; POINTI[k].z = z ;
}

/*************************************************************************************/
void All_points2vue(int npt)
{
int P;

#ifdef BYALLOC
	if (POINTP == NULL)
	{
/* Allocation buffer points projetes */
		POINTP = (type_point *) malloc ((npt+1)*sizeof(type_point)) ;
		if (POINTP == NULL) 
			exitmes ("Echec realloc POINTP") ;
	}
#endif

	/* Projection des points dans repere camera -> POINTP */
	/* et dfinition des facteurs d'chelle du trac */
    for (P = 0 ; P < npt ; P++)	
	{
		obj2vue (POINTI[P].x, POINTI[P].y, POINTI[P].z,
										&POINTP[P].x, &POINTP[P].y, &POINTP[P].z) ;
	}
}
/*************************************************************************************/
void makerover (void) ;
void makeMNT (void);
void init_rover () ;
void geodir_rover (double *, double *) ;
int geoinv_rover (double *xyzltr, double *conf, double tad, double trd,
					double E, int imax);
int makpointsRov (int) ;
int makpointsMnt (int) ;
/*************************************************************************************/
void init_allocs(void)
{
#ifdef BYALLOC
/* Allocation initiale d'un lot de points */
    POINTI = (type_point *) malloc (nbptsup*sizeof(type_point)) ;
    if (POINTI == NULL) 
		exitmes ("Echec alloc POINTI") ;
/* Allocation initiale d'un lot de triangles */
    TRIGLE = (type_trigle *) malloc (nbtrisup*sizeof(type_trigle)) ;
    if (TRIGLE == NULL)
		exitmes ("Echec alloc TRIGLE") ;
/* Allocation initiale d'un lot d'aretes */
    ARETES = (type_arete *) malloc (nbaretsup*sizeof(type_arete)) ;
    if (ARETES == NULL)
		exitmes ("Echec alloc ARETES") ;
/* Allocation initiale d'un lot d'aretes */
    TRSET = (int *) malloc (nbsetsup*sizeof(int)) ;
    if (TRSET == NULL)
		exitmes ("Echec alloc TRSET") ;
#endif
}

static char texte[320];
void fin_showrov(void)
{
int n;
char *pt = texte;
#ifdef BYALLOC
    free (POINTI) ;
    free (POINTP) ;
    free (TRIGLE) ;
    free (TRSET) ;
	free (ARETES);

    n = sprintf (pt, "Tableaux alloues en mmoire :\n") ; pt += n;
    n = sprintf (pt, "Nbr total de points (Ini+Proj) : %d/%d\n",npt_tot2+1, nbptsup) ; pt += n;
    n = sprintf (pt, "Nbr total de triangles (TRIGLE): %d/%d\n",nbtri2, nbtrisup) ; pt += n;
    n = sprintf (pt, "Nbr total de listes (TRSET)    : %d/%d\n",nbsetmax, nbsetsup) ; pt += n;
    n = sprintf (pt, "Nbr total d'aretes (ARETES)    : %d/%d\n",nbaret2, nbaretsup) ; pt += n;

#else //!BYALLOC
    n = sprintf (pt, "Tableaux statiques en mmoire :\n") ; pt += n;
    n = sprintf (pt, "Utilisation points (Ini+Proj) : %d/%d\n",npt_tot2+1, MAXPOINTS) ; pt += n;
    n = sprintf (pt, "Utilisation triangles (TRIGLE): %d/%d\n",nbtri2, MAXTRIGLES) ; pt += n;
    n = sprintf (pt, "Utilisation listes (TRSET)    : %d/%d\n",nbsetmax, MAXTRSET) ; pt += n;
    n = sprintf (pt, "Utilisation aretes (ARETES)   : %d/%d\n",nbaret2,MAXARETES) ;

#endif //BYALLOC
#ifdef VERBEUX
	MyMessBox(texte);
#endif //VERBEUX
}

void GenFasCach(int ntri, int naret)
{
int i, P, Q, k, trntr, jtr, inside;
double xP, yP, zP, xQ, yQ, zQ;
double XP, YP, XQ, YQ, Xlft, Ylft, Xrght, Yrght ;
double denom, pente ;

	if (hemf) 
	{
		DeleteEnhMetaFile(hemf);
		hemf = NULL;
	}
	hdc_cou = CreateEnhMetaFile (NULL, NULL, &RectMeta, NULL) ;

	/* Initialisation matrice pixels virtuels */
    for (k = 0 ; k < NCAR*NCAR ; k++)
    {
		ppix = &CASES[k] ;
		ppix->tr_cov = -1 ; ppix->tr_dist = DBL_MAX ;
		ppix->start = NULL ;
    }

	/* Toutes les aretes invisibles (sauf les individuelles) */
    for (i = 0 ; i < naret ; i++)
	{
		if (ARETES[i].visible >=0) ARETES[i].visible = 0;
	}

	/* Examen des triangles, activation des artes des faces visibles et mise  jour du tableau CASES */
	for (i = 0 ; i < ntri; i++) exam_trig(i);

	/* Incorpore les triangles recouvrant aux listes des  carres ecran */
    for (k = 0 ; k < NCAR*NCAR ; k++)
    {
		ppix = &CASES[k] ;
		if (ppix->tr_cov >= 0)
		{
			pnode = (type_node *) malloc (sizeof (type_node)) ;
			if (pnode == NULL) exitmes ("Echec alloc pnode #2 !") ;
			pnode->jtr = ppix->tr_cov ;
			pnode->next = ppix->start ;
			ppix->start = pnode ;
		}
	}

	/* Trace les segments des qu'ils sont visibles */
	for (k = 0 ; k < naret ; k++)
	{
		if (ARETES[k].visible ==0) continue;
		P = ARETES[k].ori;
		ppoint = &POINTP[P] ;
		xP = ppoint->x ; yP = ppoint->y ; zP = ppoint->z ;
		XP = xP / zP ; YP = yP / zP ;
		Q = ARETES[k].des;
		ppoint = &POINTP[Q] ;
		xQ = ppoint->x ; yQ = ppoint->y ; zQ = ppoint->z ;
		XQ = xQ / zQ ; YQ = yQ / zQ ;
	/* En utilisant la liste ecran, on construit l'ensemble des
	   triangles qui peuvent eventuellement cacher PQ */
	   if (XP < XQ || (XP == XQ && YP < YQ))
	   { Xlft = XP ; Ylft = YP ; Xrght = XQ ; Yrght = YQ ; } else
	   { Xlft = XQ ; Ylft = YQ ; Xrght = XP ; Yrght = YP ; }
	   ipixleft = xwhole (Xlft) ; ipixright = xwhole (Xrght) ;
	   denom = Xrght - Xlft ; if (fabs (denom) <= eps) denom = eps ;
	   pente = (Yrght-Ylft)/denom ;jbot = jtop = ywhole (Ylft) ;
	   inside = 0;
	   if (ipixleft >= 0 && ipixleft < NCAR && jbot >= 0 && jbot < NCAR) inside = 1;
	   for (ipix = ipixleft; ipix <= ipixright ; ipix++)
	   {
		   if (ipix < 0 || ipix >= NCAR) continue;
		   if (ipix == ipixright) jI = ywhole (Yrght) ;
		   else jI = ywhole (Ylft+pente*(xreal(ipix+1)-Xlft)) ;
		   if (jI >= 0 && jI < NCAR) inside = 1;
		   LOWER[ipix] = min (jbot, jI) ; jbot = jI ;
		   UPPER[ipix] = max (jtop, jI) ; jtop = jI ;
	   }
	   if (!inside) continue;
	   /* Construction de TRSET[0], ... , TRSET[nbset-1] : ensemble des
	   triangles qui peuvent cacher des points de PQ */
	   nbset = 0 ;
	   for (ipix = ipixleft; ipix <= ipixright ; ipix++)
	   {
		   if (ipix < 0 || ipix >= NCAR) continue;
		   for (jpix = LOWER[ipix]; jpix <= UPPER[ipix] ; jpix++)
		   {
				if (jpix < 0 || jpix >= NCAR) continue;
				ppix = &CASES[ipix*NCAR+jpix] ;
				pnode = ppix->start ;
				while (pnode != NULL)
				{
				   trntr = pnode->jtr ; /* # du triangle recouvrant la carr */
				   /* tntr ne sera memorise que s'il n'est pas encore
				   present dans le tableau TRSET (ensembles des trigles) */
				   TRSET[nbset] = trntr ; /* sentinelle */
				   jtr = 0 ;
				   while (TRSET[jtr] != trntr) jtr++ ;
				   if (jtr == nbset) 
				   {
					   nbset++ ; /* trntr est memorise */
#ifdef BYALLOC
					   if (nbset == nbsetsup) 
					   {
							nbsetsup *= 2 ;
							TRSET = (int *) realloc (TRSET, nbsetsup*sizeof(int)) ;
							if (TRSET == NULL)
								exitmes ("Echec realloc TRSET") ;
					   }
#endif
				   }
				   pnode = pnode->next ;
				}
			}
	   }
	   if (nbset > nbsetmax) nbsetmax = nbset; // Pour impression

	   /* Maintenant TRSET[0], ... , TRSET[nbset-1] est l'ensemble des
	   triangles qui peuvent cacher des points de PQ */
	   visibilite (xP, yP, zP, xQ, yQ, zQ, 0) ;
	}

	/* Liberation contenu des cases et connexions */
    for (k = 0 ; k < NCAR*NCAR ; k++)
    {
		type_node *ptamp ;

		pnode = CASES[k].start ;
		while (pnode != NULL)
		{ ptamp = pnode ; pnode = pnode->next ; free (ptamp) ; }
		CASES[k].start = NULL ;
    }
	hemf = CloseEnhMetaFile(hdc_cou); // Pret a etre affich
	hdc_cou = NULL;
	UpdateDraw();
} //fin GenFasCach

void init_showrov (void)
{
int i;
	/* Allocation ou dimensionnement tableaux de travail */
	init_allocs();

	/* Initialisation calculs geomtriques */
    init_rover () ;

	/* Calcul des coords de tous les POINTI (/ Inertiel) */
    npt_tot1 = makpointsRov (1) ; 
	nbtri = 0 ; nbaret = 0;
	makerover () ;
	nbtri1 = nbtri; nbaret1 = nbaret; /* memo pour rover seul */

	/* On ajoute le mnt */
	npt_tot2 = makpointsMnt (npt_tot1) ; 
	makeMNT();
	nbtri2 = nbtri; nbaret2 = nbaret; /* memo pour rover et mnt */

	i = geoinv_rover (xyzltr, conf_rov, 0., 0., 0.005, 10);
    makpointsRov (1) ; 

	
	/* dfinition d'une configuration de la camera */
	conf_cam[0] = -30. D2RAD ;		/* azimut */
	conf_cam[1] = -30. D2RAD ;		/* site   */
	conf_cam[2] = 0. D2RAD ;		/* roulis   */
	conf_cam[3] = 0.;			/* TX scene */
	conf_cam[4] = 0.;			/* TY scene */
	conf_cam[5] = 0.;			/* TZ scene */
	conf_cam[6] = 0.;			/* Tx cam */
	conf_cam[7] = 0.;			/* Ty cam */
	conf_cam[8] = 17.;			/* Tz cam */
	conf_cam[9] = TgVue;		/* gain */
	carol2mrot(matrot, conf_cam) ;
	
	/* Projection de tous les points dans le repre camra */
	All_points2vue(npt_tot2);


	/* Generation vue initiale */	
	GenFasCach(nbtri2, nbaret2) ;
}

void UpdateRov(int fmodconf, int fmodvue, int fmodsit)
{
int i;

	if (fmodvue) 
	{
		if (fmodvue == 10) 
		{
			TgVue = conf_cam[9];
			Xmin = -TgVue/2.;
			Ymin = -TgVue/2.;
			deltaX = TgVue/NCAR;
			deltaY = TgVue/NCAR;
			GenFasCach(nbtri2, nbaret2) ;
		}
		else
		{	
			carol2mrot(matrot, conf_cam) ;
			All_points2vue(npt_tot2); 
			GenFasCach(nbtri2, nbaret2) ;
		}
		return ;
	}

	if (fmodconf)
	{
		geodir_rover (xyzltr, conf_rov) ;
		makpointsRov (1) ;
		All_points2vue(npt_tot1); 
		GenFasCach(nbtri1, nbaret1) ;
		return;
	}

	if (fmodsit)
	{
		geodir_rover (xyzltr, conf_rov) ;
		i = geoinv_rover (xyzltr, conf_rov, 0., 0., 0.005, 10);
		makpointsRov (1) ;
		All_points2vue(npt_tot2); 
		GenFasCach(nbtri2, nbaret2) ;
		return;
	}

	makpointsRov (1) ;
	All_points2vue(npt_tot2); 
	GenFasCach(nbtri2, nbaret2) ;
}

