/************************************************************************/
/* FICHIER GEOUTI.C : Utilitaires gomtriques 3D                       */
/************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "geouti.h"
#define PI 3.141592653589793238
/************************************************************************
    Raz VECTEUR :       nulvec(u)

    Sorties:
	double u[3]    : Vecteur resultat  (u) = 0

*************************************************************************/
void nulvec(double* u)
    {
    u[0] = 0. ;
    u[1] = 0. ;
    u[2] = 0. ;
    }
/************************************************************************
    Recopie DE VECTEUR :        cpyvec(u,v)

    Entrees:
        double v[3]    : Vecteur

    Sorties:
        double u[3]    : Vecteur resultat  (u) = (v)

*************************************************************************/
void cpyvec(double* u, double* v)
{
    u[0] = v[0] ;
    u[1] = v[1] ;
    u[2] = v[2] ;
}
/************************************************************************
    MULTIPLICATION VECTEUR PAR UN SCALAIRE :    scavec(u, s, v)

    Entrees:
        double v[3]    : Vecteur
        double s       : scalaire

    Sorties:
        double u[3]    : Vecteur resultat  (u) = s.(v)

*************************************************************************/
void scavec(double* u, double s, double* v)
    {
    u[0] = s * v[0] ;
    u[1] = s * v[1] ;
    u[2] = s * v[2] ;
    }
/************************************************************************
    SOUSTRACTION DE VECTEUR :   subvec(u,v,w)

    Entrees:
        double v[3]    : Vecteur
        double w[3]    : Vecteur a soustraire

    Sorties:
        double u[3]    : Vecteur resultat  (u) = (v) - (w)

*************************************************************************/
void subvec(double* u, double* v, double* w)
    {
    u[0] = v[0] - w[0] ;
    u[1] = v[1] - w[1] ;
    u[2] = v[2] - w[2] ;
    }
/************************************************************************
    ADDITION DE VECTEUR :       addvec(u,v,w)

    Entrees:
        double v[3]    : Vecteur
        double w[3]    : Vecteur a additionner

    Sorties:
        double u[3]    : Vecteur resultat  (u) = (v) + (w)

*************************************************************************/
void addvec(double* u, double* v, double* w)
{
    u[0] = v[0] + w[0] ;
    u[1] = v[1] + w[1] ;
    u[2] = v[2] + w[2] ;
}

/************************************************************************
    CUMULATION DE VECTEUR :   cumvec(u,v,a,w)

    Entrees:
        double v[3]    : Vecteur depart
        double a             : scalaire
        double w[3]    : Vecteur direction

    Sorties:
        double u[3]    : Vecteur resultat  (u) = (v) + a.(w)

*************************************************************************/
void cumvec(double* u, double* v,double a, double* w)
{
    u[0] = v[0] + a*w[0] ;
    u[1] = v[1] + a*w[1] ;
    u[2] = v[2] + a*w[2] ;
}
/************************************************************************
    PRODUIT MATRICE DIRECTE * VECTEUR :   madvec(w,a,v)

    Entrees:
        double a[3][3] : Matrice
        double v[3]    : Vecteur

    Sorties:
	double w[3]    : Vecteur resultat  (w) = [A].(v)

*************************************************************************/
void madvec(double* w, double* a, double* v)
{
double p, q ;
      p  = a[0] * v[0] + a[3] * v[1] + a[6] * v[2];
      q  = a[1] * v[0] + a[4] * v[1] + a[7] * v[2];
    w[2] = a[2] * v[0] + a[5] * v[1] + a[8] * v[2];
    w[0] = p ;
    w[1] = q ;
}

/************************************************************************
    PRODUIT MATRICE DIRECTE * MATRICE DIRECTE :   madmad(a,b,c)

    Entrees:
        double b[3][3] : Matrice
	double c[3][3] : Matrice

    Sorties:
        double a[3][3] : Matrice resultat  [A] = [B].[C]

*************************************************************************/
void madmad(double* a,double* b, double* c)
{
register int i ;
double d[8] ;
    d[0] = b[0] * c[0] + b[3] * c[1] + b[6] * c[2];
    d[1] = b[1] * c[0] + b[4] * c[1] + b[7] * c[2];
    d[2] = b[2] * c[0] + b[5] * c[1] + b[8] * c[2];

    d[3] = b[0] * c[3] + b[3] * c[4] + b[6] * c[5];
    d[4] = b[1] * c[3] + b[4] * c[4] + b[7] * c[5];
    d[5] = b[2] * c[3] + b[5] * c[4] + b[8] * c[5];

    d[6] = b[0] * c[6] + b[3] * c[7] + b[6] * c[8];
    d[7] = b[1] * c[6] + b[4] * c[7] + b[7] * c[8];
    a[8] = b[2] * c[6] + b[5] * c[7] + b[8] * c[8];
    for(i = 0 ; i < 8 ; ++i) a[i] = d[i] ;
}

/************************************************************************
    PRODUIT POR DIRECT * POR DIRECT :   pordpord(a,b,c)

    Entrees:
        double b[4][3] : Por (Position -orientation)
        double c[4][3] : Por (Position -orientation)

    Sorties:
        double a[4][3] : Por resultat  [A] = [B].[C]

*************************************************************************/
void pordpord(double* a, double* b, double* c)
{
double d[3] ;
    madvec(d, b+3, c);
    madmad(a+3, b+3, c+3);
    addvec(a, b, d) ;
}
/************************************************************************
    PRODUIT POR DIRECT * VECTEUR :   pordvec(a,b,c)

    Entrees:
        double b[4][3] : Por (Position -orientation)
	double c[3]    : vecteur

    Sorties:
	double a[3]    : vecteur

*************************************************************************/
void pordvec(double* a, double* b,double* c)
{
double d[3] ;
    madvec(d, b+3, c);
    addvec(a, b, d) ;
}

/************************************************************************
    GENERATION MATRICE ROTATION :  mrotva(m,v,a)

    Entrees:
       double v[3]    : Vecteur unitaire axe rotation
       double a     : Angle rotation en radians

    Sorties:
       double m[3][3] : Matrice rotation

    Remarque: &m peut etre identique a &v

*************************************************************************/
void mrotva(double* m, double* v, double a)
{
    double wx, wy, wz;
    double k;

    wx = v[0];
    wy = v[1];
    wz = v[2];
    k = 1.-cos(a);
    m[0] = 1. + k * (wx * wx - 1.);
    m[4] = 1. + k * (wy * wy - 1.);
    m[8] = 1. + k * (wz * wz - 1.);
    m[1] = k * wx * wy;
    m[2] = k * wz * wx;
    m[5] = k * wy * wz;
    k = sin(a);
    wx = k * wx;
    wy = k * wy;
    wz = k * wz;
    m[7] = m[5] - wx;
    m[5] += wx;
    m[6] = m[2] + wy;
    m[2] -= wy;
    m[3] = m[1] - wz;
    m[1] += wz;
}

/************************************************************************
    GENERATION MATRICE ROTATION :  mrotna(m,k,a)

    Entrees:
       int k        : Numero axe de base (1, 2 ou 3)
       double g     : Angle rotation en radians

    Sorties:
       double a[3][3] : Matrice rotation

*************************************************************************/
void mrotna(double* a, int k, double g)
{
double s, c;
    c = cos(g);
    s = sin(g);
    switch (k)
    {
	case 1:
	    a[0] = 1.;
	    a[1] = a[2] = a[3] = a[6] = 0.;
	    a[4] = a[8]= c;
	    a[5] = s;
	    a[7] = -s;
	    break ;
	case 2:
	    a[4] = 1.;
	    a[1] = a[3] = a[5] = a[7] = 0.;
	    a[0] = a[8]= c;
	    a[6] = s;
	    a[2] = -s;
	    break ;
	case 3:
            a[8] = 1.;
            a[2] = a[5] = a[6] = a[7] = 0.;
            a[0] = a[4]= c;
            a[1] = s;
	    a[3] = -s;
	    break ;
    }
}

/************************************************************************
    CALCUL MATRICE POUR ANGLES L (Oz) , T (Ox1) R (Oy2).

    Entrees:
       double ltr[3] : 3 angles.

    Sorties:
       double a[9] : Matrice de rotation
       et pointeur sur la matrice.

*************************************************************************/
void ltrmat(double *a, double *ltr)
{
double cl, sl, ct, st, cr, sr ;
double slst, clst ;

    cl = cos (ltr[0]) ; ct = cos (ltr[1]) ; cr = cos (ltr[2]) ;
    sl = sin (ltr[0]) ; st = sin (ltr[1]) ; sr = sin (ltr[2]) ;
    slst = sl * st ; clst = cl * st ;

    a[0] = cl*cr - slst*sr ; a[3] = -sl*ct ; a[6] = cl*sr + slst*cr ;
    a[1] = sl*cr + clst*sr ; a[4] =  cl*ct ; a[7] = sl*sr - clst*cr ;
    a[2] =          -ct*sr ; a[5] =     st ; a[8] =           ct*cr ;
}