#include <stdlib.h>
#include <math.h>
#include <float.h>

void MyMessBox(char *);
void stkpoint (int,double,double,double) ;
void makequadri (int,int,int,int,int) ;
void maketrig (int,int,int,int) ;
void makeline(int,int) ;

typedef struct { double x,y,z ;          /* Coordonnees des points */
               } type_point ;

extern type_point *POINTI;

#define DIMMNT 30.
#define PASMNT 2.

static int FirstPL, nbinter, nbpot, nbpotp1;

int makpointsMnt (int N) 
{
int n = N;
double xini, yini, x, xx, y, yy, z;
int i, j;

	nbinter = (int) (0.5 + DIMMNT/PASMNT);
	nbpot = nbinter+1;
	nbpotp1 = nbpot+1;
	xini = -DIMMNT/2.;
	yini = -DIMMNT/2.;


	FirstPL = n;
	for (j = 0;   j < nbpot; j++)
	{
		yy = yini + j*PASMNT;
		y = yy/5.;
		for (i = 0; i < nbpot; i++)
		{
			xx = xini + i * PASMNT;
			x = xx/5.;
			z =  3.*(1-x)*(1-x)*exp(-(x*x) - (y+1)*(y+1))
			   - 10.*(x/5 - x*x*x - y*y*y*y*y)*exp(-x*x-y*y)
			   - (1./3.)*exp(-(x+1)*(x+1) - y*y);
			stkpoint (n++, xx, yy, - 0.25 + z/3.5) ;
		}
	}
	return n;
}

void makeMNT (void)
{
int i,j,n1,n;

	for (j = 0, n1 = FirstPL; j < nbinter; j++, n1 += nbpot)
	{
		for (i = 0, n = n1; i < nbinter; i++, n++)
		{
			if ((i+j)%2)
			{
				maketrig(n, n+nbpot, n+1, 1);
				maketrig(n+1, -(n+nbpot), n+nbpot+1, 1);
			}
			else
			{
				maketrig(n, n+nbpot+1, n+1, 1);
				maketrig(n+nbpot, n+nbpot+1, -n, 1);
			}
		}
	}
}


double altseg(double *Ms, double R, double *Ma, double *Mb)
/*
% Entres :
%   Ms        : Centre de la spre
%   R         : rayon de la spre
%   Ma        : Point A du segment
%   Mb        : Point B du segment
% Sorties :
%   dZ :  l'altitude du point P de la sphre/ au segment AB
%         = inf si pas de contact ou contact hors segment
*/
{
double dZ, U2[3], n2D2, Lab2, Lab, u2[3], V[3], v2, lav2, lambda, d2;
double Discrim, c, mU1[3], SA[3], mu1, mu2, eta1, eta2, t2;
int i;

	dZ = DBL_MAX;
	for (i=0;i<3;i++) U2[i] = Mb[i]-Ma[i];
	n2D2 = U2[0]*U2[0]+U2[1]*U2[1];
	Lab2 = n2D2 + U2[2]*U2[2];
	Lab = sqrt(Lab2);
	for (i=0;i<3;i++) u2[i] = U2[i]/Lab;
	V[0] = - u2[1]; V[1] = u2[0]; V[2] = 0.;
	v2 = V[0]*V[0]+V[1]*V[1];
	for (i=0;i<3;i++) SA[i] = Ma[i]-Ms[i];
	lav2 = V[0]*SA[0]+V[1]*SA[1];
	lambda = lav2/v2;
	d2 = lambda*lav2;
	Discrim = R*R - d2;
	if (Discrim < 0.) return dZ;  // Pas de contact possible
	c = u2[2];
	for (i=0;i<3;i++) mU1[i] = - c*u2[i];
	mU1[2] += 1. ;
	mu1 = (SA[0]*mU1[0] + SA[1]*mU1[1] + SA[2]*mU1[2])/v2;
	mu2 = (-SA[0]*u2[0] - SA[1]*u2[1] + SA[2]*(c-u2[2]))/v2;
	eta1 = sqrt(Discrim/v2);
	eta2 = c*eta1;
	t2 = mu2 + eta2;
	if (t2 < 0 || t2 > Lab) return dZ; // Contact a l''exterieur de AB
	dZ = -(eta1+mu1);
	return dZ;
}

double altrig(double *Ms, double R, double *Mc, double Za, double Zb, int sx, int sy, double D)
/*
% Entres :
%   Ms        : Centre de la spre
%   R         : rayon de la spre
%   Mc        : sommet C du triangle rectangle,
%   Za        : altitude du sommet A situ  un X diffrent
%   Zb        : altitude du sommet B situ  un Y diffrent
%   sx        : +/- 1 suivant que A est du cot des x >0 ou <0
%   sy        : +/- 1 suivant que B est du cot des y >0 ou <0
%   D         : cart en X ou Y entre A et C ou B et C
% Sorties :
%   dZ    :  l'altitude du point P de la sphre/ au plan ABC
%            = inf si pas de contact ou contact hors triangle
*/
{
double dZ, a, b, r, Xp, Xe, sXe, Yp, Ye, sYe;

	dZ = DBL_MAX;
	a = sx*(Mc[2]-Za);
	b = sy*(Mc[2]-Zb);
	r = R/sqrt(a*a + b*b + D*D);
	Xp = Ms[0] - r*a;
	Xe = Xp - Mc[0]; 
	sXe = sx*Xe/D;
	if (sXe < 0.) return dZ;
	Yp = Ms[1] -r*b;
	Ye = Yp - Mc[1]; 
	sYe = sy * Ye/D;
	if (sYe < 0.) return dZ;
	if (sXe + sYe > 1.) return dZ;
	dZ = Ms[2] - r*D - Mc[2] + (a*Xe + b*Ye)/D;
	return dZ;
}

static int Iytl[8][3]  = {{0,0,1},{0,0,1},{1,1,0},{1,1,2},{2,2,1},{2,2,1},{1,1,2},{1,1,0}};
static int Ixtl[8][3]  = {{0,1,0},{2,1,2},{1,2,1},{1,2,1},{2,1,2},{0,1,0},{1,0,1},{1,0,1}};
static int lsegl[8][4] = {{1,0,0,0},{3,0,0,0},{2,3,4,9},{4,5,6,9},{5,0,0,0},{7,0,0,0},{6,7,8,9},{8,1,2,9}};
static int Iysl[8][2]  = {{1,0},{1,0},{0,1},{1,1},{1,2},{1,2},{2,1},{1,1}};
static int Ixsl[8][2]  = {{0,1},{1,1},{1,2},{1,2},{2,1},{1,1},{1,0},{1,0}};
static int Iyte[8][3]  = {{0,0,1},{0,0,1},{1,1,0},{1,1,2},{2,2,1},{2,2,1},{1,1,2},{1,1,0}};
static int Ixte[8][3]  = {{1,0,1},{1,2,1},{2,1,2},{2,1,2},{1,2,1},{1,0,1},{0,1,0},{0,1,0}};
static int lsege[8][4] = {{1,2,9,0},{2,3,9,0},{3,4,9,0},{4,5,9,0},{5,6,9,0},{6,7,9,0},{7,8,9,0},{8,1,9,0}};
static int Iyse[8][2]  = {{1,0},{1,0},{1,0},{1,1},{1,2},{1,2},{1,2},{1,1}};
static int Ixse[8][2]  = {{1,0},{1,1},{1,2},{1,2},{1,2},{1,1},{1,0},{1,0}};

double h2sol(double *Pos, double R)
/*
% Entree :
%   Pos      : Position centre S de la roue
%   R       : Rayon de la roue
%   MNT     : Matrice de altitudes (iy,ix)
%   D       : Taille de la maille
% Sortie :
%   DeltaH  : Hauteur du point de contact potentiel P au dessus du terrain
%   SP      : Vecteur centre-Point de contact potentiel (dans Ro)
%   Uk      : Sensibilit de DeltaH  la position de S
%
*/
{
double D = PASMNT, deltaH;
int Nx = nbpot, Ny = nbpot, N2 = Nx*Ny, nx, ny, nx0, ny0;
double x,y,u,v, X,Y,Z, X0, Y0, X1, Y1, dX, dY, Delta, Zp;
double Zyx[3][3], Myx[3][3][3], dz;
int iy, ix, iiy, iix;
int ETOILE;
int (*Ixt)[3], (*Iyt)[3], (*lseg)[4], (*Ixs)[2],(*Iys)[2] ;
int i, k, kk, sx, sy, fseg[9];
int iyy;
double Ms[3];

	Ms[0] = Pos[0]+DIMMNT/2.; Ms[1] = Pos[1]+DIMMNT/2.; Ms[2] = Pos[2];

	// Dtermination du coin origine des 4 carrs englobant la sphre
	// (ny0,nX0)
	//
	x = Ms[0]; nx = (int) (x/D); u = x/D-nx;
	if (u <= 0.5) nx0 = nx-1; else nx0 = nx; 
	X0 = nx0*D; 
	y = Ms[1]; ny = (int) (y/D); v = y/D-ny;
	if (v <= 0.5) ny0 = ny-1; else ny0 = ny;
	Y0 = ny0*D; 

	// Extraction des coordonnes des 9 points du maillage
	Y = Y0;
	for (iy = 0; iy < 3; iy++)
	{
		X = X0;
		iiy = ny0+iy;
		iyy = iiy*Nx;
		for (ix = 0; ix < 3; ix++)
		{
			iix = iyy + nx0 + ix;
			if (iix < 0 || iix >= N2) Z = 0.;
			else Z = POINTI[FirstPL+ iix].z;
			Zyx[ix][iy] = Z;
			Myx[ix][iy][0] = X;
			Myx[ix][iy][1] = Y;
			Myx[ix][iy][2] = Z;
			X = X + D;
		}
		Y = Y + D;
	}

	// Extraction des indices des sommets des triangles et segments
	if ((nx0+ny0)%2)
	{
		ETOILE = 0;
		Iyt = Iytl; Ixt = Ixtl; lseg = lsegl;
		Iys = Iysl; Ixs = Ixsl;
	}
	else
	{
		ETOILE = 1;
		Iyt = Iyte; Ixt = Ixte; lseg = lsege;
		Iys = Iyse; Ixs = Ixse;
	}

	for (k = 0; k <= 8; k++) fseg[k] = 1; // Flag pour utilit traitement segment(0:7) et sommet(8)
	// Recherche contact sur les 8 triangles
	deltaH = DBL_MAX;
	for (k = 0; k < 8; k++)
	{
		fseg[k] = 1;
		sx = Ixt[k][1] - Ixt[k][0];
		sy = Iyt[k][2] - Iyt[k][0];
		dz = altrig(Ms, R, Myx[Ixt[k][0]][Iyt[k][0]],
			Zyx[Ixt[k][1]][Iyt[k][1]], Zyx[Ixt[k][2]][Iyt[k][2]], sx, sy, D);
		if (dz < DBL_MAX)
		{
		  if (ETOILE)
		  {
			  fseg[lseg[k][0]-1]=0; fseg[lseg[k][1]-1]=0; fseg[8]=0;
		  }
		  else
		  {
			  for (i=0; i< 4; i++)
			  {
				  if (!lseg[k][i]) break;
				  fseg[lseg[k][i]] = 0;
			  }
		  }
		  if (dz < deltaH) deltaH = dz; 
		} 
	}
	// Recherche contact sur les 8 segments
	for (k = 0; k < 8; k++)
	{
	  if (fseg[k])
	  {
		kk = k+8;
		dz = altseg(Ms, R, Myx[Ixs[k][0]][Iys[k][0]], Myx[Ixs[k][1]][Iys[k][1]]);    
		if (dz < DBL_MAX)
		{
		  if (ETOILE) fseg[8]=0; 
		  else if (k%2) fseg[8]=0;
		  if (dz < deltaH)  deltaH = dz; 
		}
	  } 
	}
	if (fseg[8] == 0) return deltaH;
	// On teste le point central
	X1 = X0+D; Y1 = Y0+D;
	dX = Ms[0]-X1; dY = Ms[1]-Y1;
	Delta = R*R-(dX*dX+dY*dY);
	if (Delta < 0.)
	{
		if (deltaH == DBL_MAX)
		{
			//MyMessBox("zMNTrig : Pas de solution. Le maillage doit etre trop petit");
			//exit(1);
			return 0.;
		}
		return deltaH;
	}
	Zp = Ms[2] - sqrt(Delta);
	dz = Zp - Zyx[1][1];
	if (deltaH < dz) return deltaH;
	deltaH = dz;
	return deltaH;
}

